import enum
from typing import Dict, List, Optional

from .util_functions import enum_string


class DifficultyType(enum.Enum):
    # how much of the circuit space should be used up, values in [0, 1]
    CircuitExuberance = ("Circuit Exuberance", 1.15)
    # how many of the available qubits should be rotated, values in [0, 1]
    QubitExuberance = ("Qubit Exuberance", 1.15)
    # how many changed qubits should receive a second rotation, values in [0, 1]
    RotationExuberance = ("Rotation Exuberance", 0.7)
    # how many angles are available to rotations, values are any integers (negatives imply unrestricted real angles)
    RandomizationDegree = ("Randomization Degree", 0.9)
    # how many additional edits are given based on used gates, values >= 0
    BonusEditRatio = ("Bonus Edit Ratio", 1.1)

    # how many gates are at least needed to pick from during stv generation
    MinAvailableGates = "Minimum Available Gates", 1
    # how many unique gates are at least needed to pick form during stv generation
    MinGateVariety = "Minimum Gate Variety", 1
    # minimum difficulty sum among all available gates for stv generation
    MinGateDifficulty = "Minimum Gate Difficulty Sum", 1

    def __init__(self, name: str, level_ratio: float):
        self.__name = name
        self.__level_ratio = level_ratio  # how much this difficulty type affects the overall difficulty level

    @property
    def name(self) -> str:
        return self.__name

    @property
    def _level_ratio(self) -> float:
        return self.__level_ratio

    def __str__(self) -> str:
        return self.__name


class StvDifficulty:
    """This class is immutable."""

    __DIFF_VALUES: Dict[DifficultyType, List[int]] = {
        # during generation
        DifficultyType.CircuitExuberance:   [0.7, 0.7, 0.75, 0.8, 0.85, 0.9, 1.0],
        DifficultyType.QubitExuberance:     [0.5, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
        DifficultyType.RotationExuberance:  [0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
        DifficultyType.RandomizationDegree: [  2,  2,   4,  10,  19,  37,  0],
        DifficultyType.BonusEditRatio:      [5.0, 4.0, 3.3, 2.5, 1.6, 0.9, 0.3],
        # pre generation (i.e., preparation)
        DifficultyType.MinAvailableGates:   [  1,   1,   3,   4,   5,   6,   8],
        DifficultyType.MinGateVariety:      [  1,   1,   2,   3,   4,   5,   7],
        DifficultyType.MinGateDifficulty:   [  0,   1,   5,  10,  15,  20,  30],
    }

    @staticmethod
    def _get_diff_value(diff_type: DifficultyType, level: int) -> float:
        index = level - StvDifficulty.min_difficulty_level()   # normalize level to index
        return StvDifficulty.__DIFF_VALUES[diff_type][index]

    @staticmethod
    def min_difficulty_level() -> int:
        return 0

    @staticmethod
    def max_difficulty_level() -> int:
        return StvDifficulty.num_of_difficulty_levels() - 1

    @staticmethod
    def num_of_difficulty_levels() -> int:
        return len(StvDifficulty.__DIFF_VALUES[DifficultyType.CircuitExuberance])

    @staticmethod
    def degrees_of_freedom() -> int:
        return len(StvDifficulty.__DIFF_VALUES)

    @staticmethod
    def code_len() -> int:
        return StvDifficulty.degrees_of_freedom() * len(str(StvDifficulty.max_difficulty_level()))

    @staticmethod
    def validate() -> bool:
        test_diff = StvDifficulty.from_difficulty_code("1")
        for diff_type, values in StvDifficulty.__DIFF_VALUES.items():
            if len(values) != StvDifficulty.num_of_difficulty_levels():
                return False
            if test_diff.get_absolute_value(diff_type, 1, 1) is None:
                return False    # a conversion is missing
        return True

    @staticmethod
    def _calc_avg_level(values: Dict[DifficultyType, int]) -> int:
        level_sum = sum([values[diff_type] * diff_type._level_ratio for diff_type in values])
        return int(level_sum / len(values))

    @staticmethod
    def _compute_absolute_value(diff_type: DifficultyType, diff_dict: Dict[DifficultyType, float], num_of_qubits: int,
                                circuit_space: int, fallback_value: Optional[float]) -> int:
        if diff_type in diff_dict:
            rel_val = diff_dict[diff_type]
        else:
            rel_val = fallback_value

        if diff_type is DifficultyType.CircuitExuberance:
            return int(rel_val * circuit_space)
        if diff_type is DifficultyType.QubitExuberance:
            return int(rel_val * num_of_qubits)
        if diff_type is DifficultyType.RotationExuberance:
            return int(rel_val * StvDifficulty._compute_absolute_value(DifficultyType.QubitExuberance, diff_dict,
                                                                       num_of_qubits, circuit_space, fallback_value))
        if diff_type is DifficultyType.RandomizationDegree:
            return int(rel_val)
        if diff_type is DifficultyType.BonusEditRatio:
            return int(rel_val * StvDifficulty._compute_absolute_value(DifficultyType.CircuitExuberance, diff_dict,
                                                                       num_of_qubits, circuit_space, fallback_value))

        # the following DifficultyTypes only have absolute values, so no need for additional computations
        if diff_type in [DifficultyType.MinAvailableGates, DifficultyType.MinGateVariety,
                         DifficultyType.MinGateDifficulty]:
            return int(diff_dict[diff_type])
        raise NotImplementedError(f"No absolute value computation implemented for {enum_string(diff_type)}")

    """
    circuit_exuberance: how full (]0, 1]) the circuit should be, e.g., 0.5 = half of the circuit_space is
        used, while 1.0 implies all of it is used (if possible based on gates actually available to create the
        puzzle)
    qubit_exuberance: how many of the qubits  (]0, 1]) should be rotated randomly, e.g., 0.5 rotates half of
        all qubits, while 1.0 rotates all of them once
    rotation_exuberance: how many rotated qubits ([0, 1]) should be rotated along a second axis
    randomization_degree: how many angles are available to random rotations (2*PI / randomization_degree),
        a value of 0 implies no restrictions (i.e., any real number is allowed)
    bonus_edit_ratio: how many additional edits the puzzle provides relative to the number of used gates,
        e.g., a value of 1.0 means that the player has twice as many edits as gates actually needed to solve the
        puzzle, needs to be >= 0
    """

    @staticmethod
    def from_difficulty_code(code: str) -> "StvDifficulty":
        level_len = len(str(StvDifficulty.max_difficulty_level()))  # how many characters a single level code has
        if len(code) == level_len:
            code = code * StvDifficulty.degrees_of_freedom()   # extend code so every DiffType has the same level

        values: Dict[DifficultyType, int] = {}
        for i, diff_type in enumerate(DifficultyType):
            if i >= len(code):
                cur_level = StvDifficulty._calc_avg_level(values)   # extend by average if code is too short
            else:
                cur_level = int(code[i * level_len:(i+1) * level_len])
            values[diff_type] = cur_level
        return StvDifficulty(values)

    @staticmethod
    def from_difficulty_level(level: int) -> "StvDifficulty":
        assert StvDifficulty.min_difficulty_level() <= level <= StvDifficulty.max_difficulty_level(), "Invalid level: " \
            f"{StvDifficulty.min_difficulty_level()} <= {level} <= {StvDifficulty.max_difficulty_level()} is False!"
        return StvDifficulty.from_difficulty_code(str(level))

    def __init__(self, level_values: Dict[DifficultyType, int]):
        self.__level_values = level_values.copy()
        self.__level = StvDifficulty._calc_avg_level(level_values)

    @property
    def level(self) -> int:
        return self.__level

    def get_level(self, diff_type: DifficultyType) -> int:
        if diff_type in self.__level_values:
            return self.__level_values[diff_type]
        # return the general level if no specific level exists for diff_type
        return self.__level

    def get_relative_value(self, diff_type: DifficultyType) -> float:
        level = self.get_level(diff_type)
        assert StvDifficulty.min_difficulty_level() <= level <= StvDifficulty.max_difficulty_level(), \
            "Invalid difficulty level: " \
            f"{StvDifficulty.min_difficulty_level()} <= {level} <= {StvDifficulty.max_difficulty_level()} is False!"
        return StvDifficulty._get_diff_value(diff_type, level)

    def get_absolute_value(self, diff_type: DifficultyType, num_of_qubits: int, circuit_space: int) -> int:
        diff_dict = {}
        for dt in DifficultyType:
            diff_dict[dt] = self.get_relative_value(dt)
        fallback_value = StvDifficulty._get_diff_value(diff_type, self.level)
        return StvDifficulty._compute_absolute_value(diff_type, diff_dict, num_of_qubits, circuit_space, fallback_value)

    def get_absolute_dict(self, num_of_qubits: int, circuit_space: int) -> Dict[DifficultyType, int]:
        # 1) prepare all relative values
        rel_diff_dict = {}
        for diff_type in DifficultyType: rel_diff_dict[diff_type] = self.get_relative_value(diff_type)
        # 2) prepare the absolute values based on the relative ones
        abs_diff_dict = {}
        for diff_type in DifficultyType:
            fallback_value = StvDifficulty._get_diff_value(diff_type, self.level)
            abs_diff_dict[diff_type] = StvDifficulty._compute_absolute_value(diff_type, rel_diff_dict, num_of_qubits,
                                                                             circuit_space, fallback_value)
        return abs_diff_dict

    def create_relative(self, add_levels: Dict[DifficultyType, int]) -> "StvDifficulty":
        level_values = self.__level_values.copy()
        for diff_type in add_levels:
            new_level = level_values[diff_type] + add_levels[diff_type]
            level_values[diff_type] = max(min(new_level, StvDifficulty.max_difficulty_level()),
                                          StvDifficulty.min_difficulty_level())
        return StvDifficulty(level_values)

    def to_code(self) -> str:
        return "".join([str(self.get_level(diff_type)) for diff_type in DifficultyType])

    def __str__(self) -> str:
        return f"StvDifficulty({self.to_code()})"



from typing import List, Optional

from qrogue.game.logic.collectibles import Collectible, instruction as gates
from qrogue.util import Logger, MyRandom


class CollectibleFactory:
    @staticmethod
    def empty() -> "CollectibleFactory":
        return CollectibleFactory([None])

    def __init__(self, pool: List[Collectible]) -> None:
        if pool is None or len(pool) < 1:
            Logger.instance().throw(Exception(f"invalid pool for CollectibleFactory: {pool}"))
        self.__pool = pool.copy()
        self.__order_index = -1

    @property
    def pool_copy(self) -> List[Collectible]:  # todo make protected instead of public?
        return self.__pool.copy()

    def __produce(self, rm: Optional[MyRandom], remove: bool = False) -> Optional[Collectible]:
        if len(self.__pool) <= 0:
            return None

        if rm:
            return rm.get_element(self.__pool, remove=remove, msg="CollectibleFactory.__produce")
        else:
            self.__order_index += 1
            if self.__order_index >= len(self.__pool):
                self.__order_index = 0
            elem = self.__pool[self.__order_index]
            if remove:
                self.__pool.remove(elem)
            return elem

    def produce(self, rm: Optional[MyRandom] = None) -> Optional[Collectible]:
        return self.__produce(rm, remove=False)

    def produce_multiple(self, rm: Optional[MyRandom], num_of_elements: int, unique_elements: bool = True) \
            -> List[Collectible]:
        """
        Produces a list with multiple Collectibles. If unique_elements is True and num_of_elements is greater than the
        number of Collectibles available in the factory's pool, we start over after depleting the pool. In any case if
        the pool doesn't contain any Collectible, an empty list is returned.

        :param rm: MyRandom for random selection from pool if wanted, None implies ordered selection
        :param num_of_elements: how many Collectibles we want to produce
        :param unique_elements: whether the selected Collectibles should be unique (if possible)
        :return: a list with the specified number of elements or an empty list if the factory cannot produce
        Collectibles
        """
        if len(self.__pool) <= 0:
            return []

        if unique_elements:
            temp = self.pool_copy
            elements = []
            while len(elements) < num_of_elements:
                elem = self.__produce(rm, remove=True)
                if elem is not None:
                    elements.append(elem)
                    if len(self.__pool) <= 0:
                        self.__pool = temp.copy()
            self.__pool = temp
            return elements
        else:
            return [self.__produce(rm, remove=False) for _ in range(num_of_elements)]


class OrderedCollectibleFactory(CollectibleFactory):
    """
    Same as CollectibleFactory but instead of defaulting to random production this
    class defaults to ordered production
    """

    @staticmethod
    def from_factory(factory: CollectibleFactory):
        return OrderedCollectibleFactory(factory.pool_copy)

    def __init__(self, pool: List[Collectible]):
        super().__init__(pool)

    def produce(self, rm: Optional[MyRandom] = None) -> Collectible:
        return super(OrderedCollectibleFactory, self).produce()

    def produce_multiple(self, rm: Optional[MyRandom], num_of_elements: int, unique_elements: bool = True) \
            -> List[Collectible]:
        return super(OrderedCollectibleFactory, self).produce_multiple(None, num_of_elements, unique_elements)


class GateFactory:
    @staticmethod
    def default() -> CollectibleFactory:
        return CollectibleFactory(pool=[
            gates.HGate(), gates.XGate(),  # gates.YGate(), gates.ZGate(),
            gates.CXGate(), gates.SwapGate(),
        ])

    @staticmethod
    def quantum() -> CollectibleFactory:
        return CollectibleFactory(pool=[
            gates.HGate(), gates.SGate(), gates.YGate(), gates.ZGate(),
        ])



from datetime import datetime
from typing import Dict, List, Optional, Callable

from .achievements import Unlocks
from .config import ExpeditionConfig, MapConfig, GameplayConfig, PathConfig, MapGrammarConfig, ScoreConfig, GateType
from .stv_difficulty import StvDifficulty
from .util_functions import datetime2str


class LevelInfo:
    __LAST_LEVEL_NUM = 6

    __MAP_ORDER: Dict[int, Dict[str, str]] = {
        # map names:
        #   - the first character determines if it's a level ("l") or world ("w")
        #   - the second character determines to which world the map belongs to
        #   - the third character determines if the level differs based on knowledge mode, followed by the corresponding
        #     digit of the knowledge mode
        #   - last digit for levels is used to order the levels (only for structure, not used in game logic)
        #   - alternatively maps can also start with "expedition" to mark them as generated
        0: {
            MapConfig.first_uncleared(): "l0k0v0",
            "l0k0v6": f"{MapConfig.expedition_map_prefix()}",
        },
        1: {
            MapConfig.first_uncleared(): "l0k1v0",
            "l0k1v6": f"{MapConfig.expedition_map_prefix()}",
        },
    }

    __LEVEL_COMPLETION_UNLOCKS: Dict[str, List[Unlocks]] = {
        # newbie tutorials
        "l0k0v0": [Unlocks.MainMenuContinue, Unlocks.ShowEnergy, ],
        "l0k0v1": [Unlocks.ShowEquation, Unlocks.PuzzleHistory],
        "l0k0v4": [Unlocks.CircuitReset, Unlocks.Boss],
        "l0k0v5": [Unlocks.LevelSelection, Unlocks.MiniBoss],
        "l0k0v6": [Unlocks.Workbench, Unlocks.Expeditions],

        # experienced tutorials are copied from newbie tutorials entered in init()

        # other levels
    }

    __LEVEL_COMPLETION_UNLOCKED_GATES: Dict[str, List[GateType]] = {
        # newbie tutorials
        "l0k0v0": [GateType.XGate],
        "l0k0v1": [GateType.CXGate],
        "l0k0v2": [],
        "l0k0v3": [GateType.HGate],
        "l0k0v4": [],
        "l0kv05": [GateType.SGate],
        "l0kv06": [GateType.YGate],

        # experienced tutorials are copied from newbie tutorials entered in init()

        # other levels
    }

    __LEVEL_START_GATES: Dict[str, List[GateType]] = {      # todo: parse from level file instead?
        # newbie tutorials
        "l0k0v0": [GateType.XGate],
        "l0k0v1": [GateType.XGate],
        "l0k0v2": [GateType.XGate],
        "l0k0v3": [GateType.XGate, GateType.CXGate],
        "l0k0v4": [GateType.XGate, GateType.CXGate, GateType.HGate],
        "l0k0v5": [GateType.XGate, GateType.CXGate, GateType.HGate],
        "l0k0v6": [GateType.XGate, GateType.CXGate, GateType.HGate, GateType.YGate],

        # experienced tutorials are copied from newbie tutorials entered in init()

        # other levels
    }

    # for converting internal names to display names and vice versa
    __NAME_CONVERTER: Dict[str, str] = {}  # is filled dynamically within LevelInfo.init()

    @staticmethod
    def init():
        # initialize map order in-between tutorials
        # the last level will lead to an expedition, so it's okay that i will never be equal to __LAST_LEVEL_NUM
        for i in range(0, LevelInfo.__LAST_LEVEL_NUM):
            for km in [0, 1]:
                src_name, dst_name = f"l0k{km}v{i}", f"l0k{km}v{i + 1}"
                LevelInfo.__MAP_ORDER[km][src_name] = dst_name

        def init_experienced_tutorials(target_dict: Dict[str, List]):
            values_to_add: Dict[str, List] = {}
            for name in target_dict:
                if name[2:4] != "k0": continue
                exp_name = name[:2] + "k1" + name[4:]
                values_to_add[exp_name] = target_dict[name].copy()
            for name in values_to_add:
                target_dict[name] = values_to_add[name]

        # initialize completion unlocks for experienced tutorials
        init_experienced_tutorials(LevelInfo.__LEVEL_COMPLETION_UNLOCKS)

        # initialize gate unlocks for experienced tutorials
        init_experienced_tutorials(LevelInfo.__LEVEL_COMPLETION_UNLOCKED_GATES)

        # initialize level start gates for experienced tutorials
        init_experienced_tutorials(LevelInfo.__LEVEL_START_GATES)

        # initialize __NAME_CONVERTER
        for mode in LevelInfo.__MAP_ORDER.keys():
            for level in LevelInfo.__MAP_ORDER[mode]:
                if level == MapConfig.first_uncleared(): continue

                data = PathConfig.read_level(level, in_dungeon_folder=True)
                start = data.index(MapGrammarConfig.name_prefix()) + len(MapGrammarConfig.name_prefix())
                end = data.index(MapGrammarConfig.description_prefix()) - 1  # subtract the new line before description
                display_name = data[start + 1:end - 1]  # also remove the enclosing quotes

                LevelInfo.__NAME_CONVERTER[level] = display_name

    @staticmethod
    def extract_knowledge_mode(cur_map: str) -> Optional[int]:
        if "k" in cur_map and "v" in cur_map:
            km = cur_map[cur_map.index("k") + 1:cur_map.index("v")]
            if km.isdigit():
                return int(km)
            else:
                from qrogue.util import Config
                Config.check_reachability("LevelInfo.extract_knowledge_mode(): invalid knowledge mode")
        return None

    @staticmethod
    def get_next(cur_map: str, is_level_completed: Callable[[str], bool]) -> Optional[str]:
        """
        :param cur_map: internal name of the map we want to get the next map's internal name for
        :param is_level_completed: Callable to find out which levels are already completed in case cur_map equals the
                                    meta name MapConfig.first_uncleared()
        :return: the internal name of cur_map's next map or None if there is no next map
        """
        knowledge_mode = LevelInfo.extract_knowledge_mode(cur_map)
        if knowledge_mode is None:
            knowledge_mode = GameplayConfig.get_knowledge_mode()

        if cur_map == MapConfig.first_uncleared():
            next_map = LevelInfo.__MAP_ORDER[knowledge_mode][cur_map]
            while is_level_completed(next_map):
                if next_map in LevelInfo.__MAP_ORDER[knowledge_mode]:
                    next_map = LevelInfo.__MAP_ORDER[knowledge_mode][next_map]
                else:
                    break
            return next_map
        elif cur_map in LevelInfo.__MAP_ORDER[knowledge_mode]:
            return LevelInfo.__MAP_ORDER[knowledge_mode][cur_map]
        elif cur_map.startswith(MapConfig.expedition_map_prefix()):
            return MapConfig.expedition_map_prefix()
        return None

    @staticmethod
    def get_prev(cur_map: str, is_level_completed: Callable[[str], bool]) -> Optional[str]:
        """
        :param cur_map: internal name of the map we want to get the previous map's internal name for
        :param is_level_completed: Callable to find out which levels are already completed in case cur_map equals the
                                    meta name MapConfig.first_uncleared()
        :return: the internal name of cur_map's previous map or None if there is no previous map
        """
        if cur_map == MapConfig.first_uncleared():
            cur_map = LevelInfo.get_prev(LevelInfo.get_next(cur_map, is_level_completed), is_level_completed)
        knowledge_mode = LevelInfo.extract_knowledge_mode(cur_map)
        if knowledge_mode is None:
            knowledge_mode = GameplayConfig.get_knowledge_mode()

        if cur_map.startswith(MapConfig.expedition_map_prefix()):
            return f"l0k{knowledge_mode}v4"  # last level
        else:
            level = cur_map[cur_map.index("v") + 1:]
            if level.isnumeric() and level != "0":
                level = int(level) - 1
            else:
                return None
            return f"l0k{knowledge_mode}v{level}"

    @staticmethod
    def get_level_completion_unlocks(level_name: str, is_level_completed: Callable[[str], bool],
                                     include_previous_levels: bool = False) -> List[Unlocks]:
        if include_previous_levels:
            unlocks = []
            while level_name is not None:
                unlocks += LevelInfo.get_level_completion_unlocks(level_name, is_level_completed,
                                                                  include_previous_levels=False)
                level_name = LevelInfo.get_prev(level_name, is_level_completed)
            return unlocks
        else:
            if level_name in LevelInfo.__LEVEL_COMPLETION_UNLOCKS:
                return LevelInfo.__LEVEL_COMPLETION_UNLOCKS[level_name]
            return []

    @staticmethod
    def get_level_completion_unlocked_gates(level_name: str, is_level_completed: Callable[[str], bool],
                                            include_previous_levels: bool = False) -> List[GateType]:
        if include_previous_levels:
            gates = []
            while level_name is not None:
                gates += LevelInfo.get_level_completion_unlocked_gates(level_name, is_level_completed,
                                                                       include_previous_levels=False)
                level_name = LevelInfo.get_prev(level_name, is_level_completed)
            return gates
        else:
            if level_name in LevelInfo.__LEVEL_COMPLETION_UNLOCKED_GATES:
                return LevelInfo.__LEVEL_COMPLETION_UNLOCKED_GATES[level_name]
            return []

    @staticmethod
    def get_level_start_gates(level_name: str) -> List[GateType]:
        """
        Provides a list of GateType based on which gates the player has when starting a level without special settings
        (i.e., on the first play-through).

        :param level_name: internal name of the level we want to get the starting gates for
        :return: list of GateType corresponding to the gates the player has as the beginning of the referenced level
        """
        if level_name in LevelInfo.__LEVEL_START_GATES:
            return LevelInfo.__LEVEL_START_GATES[level_name]
        return []

    @staticmethod
    def convert_to_display_name(internal_name: str, allow_display_name: bool = True) -> Optional[str]:
        """
        Returns: the display name corresponding to the provided level name or None if no level name was provided
        Args:
            internal_name: the internal name of the level we want to retrieve the display name of
            allow_display_name: whether we allow internal_name to already be a display name and return it in case it is
        """
        if internal_name.startswith(MapConfig.expedition_map_prefix()):
            from qrogue.util import Config
            Config.check_reachability(f"LevelInfo.convert_to_display_name({internal_name})")
            return f"Expedition"

        if internal_name in LevelInfo.__NAME_CONVERTER:
            return LevelInfo.__NAME_CONVERTER[internal_name]

        if allow_display_name:
            # this call cannot allow an internal name to avoid an endless recursion
            return LevelInfo.convert_to_internal_name(internal_name, allow_internal_name=False)
        return None

    @staticmethod
    def convert_to_internal_name(display_name: str, allow_internal_name: bool = True) -> Optional[str]:
        """
        Returns: the internal name corresponding to the provided level name or None if no level name was provided
        Args:
            display_name: the display name of the level we want to retrieve the internal name of
            allow_internal_name: whether we allow display_name to already be an internal name and return it in that case
        """
        for internal in LevelInfo.__NAME_CONVERTER:
            if LevelInfo.__NAME_CONVERTER[internal] == display_name:
                return internal

        if allow_internal_name:
            # this call cannot allow a display name to avoid an endless recursion
            return LevelInfo.convert_to_display_name(display_name, allow_display_name=False)
        return None

    @staticmethod
    def get_expedition_difficulty(expedition_progress: int) -> int:
        # level is at least MIN_DIFFICULTY and at most max_difficulty_level
        return min(max(int(expedition_progress / 10), ExpeditionConfig.MIN_DIFFICULTY),
                   StvDifficulty.max_difficulty_level())

    @staticmethod
    def get_default_num_of_qubits(difficulty_level: int) -> int:
        if difficulty_level <= 3:
            return 2
        return 3

    @staticmethod
    def get_default_circuit_space(difficulty_level: int) -> int:
        if difficulty_level <= 3:
            return 5
        elif difficulty_level <= 4:
            return 6
        return 7


class LevelData:
    def __init__(self, name: str, date_time: datetime, duration: int, score: int):
        if name.endswith("done"):
            self.__name = name[:-len("done")]
        else:
            self.__name = name
        self.__date_time = date_time
        self.__duration = duration
        self.__score = score
        self.__time_bonus = ScoreConfig.compute_time_bonus(score, duration)

    @property
    def name(self) -> str:
        """

        Returns: internal name of the map

        """
        return self.__name

    @property
    def is_lesson(self) -> bool:
        return self.__name.startswith(MapConfig.lesson_map_prefix())

    @property
    def date_time(self) -> datetime:
        return self.__date_time

    @property
    def duration(self) -> int:
        return self.__duration

    @property
    def score(self) -> int:
        return self.__score

    @property
    def time_bonus(self) -> int:
        return self.__time_bonus

    @property
    def total_score(self) -> int:
        return self.__score + self.__time_bonus

    @property
    def knowledge_mode(self) -> Optional[int]:
        return LevelInfo.extract_knowledge_mode(self.__name)

    @property
    def level_num(self) -> Optional[int]:
        if "v" in self.__name:
            level_num = self.__name[self.__name.index("v")+1:]
            if level_num.isdigit():
                return int(level_num)
        return None

    def __str__(self) -> str:
        return f"{self.__name} ({datetime2str(self.__date_time)}, {self.__duration}s, " \
               f"#{self.total_score})"



from .config import Config
from .error_config import ErrorConfig
from .gameplay_config import CheatConfig, GameplayConfig
from .gameplay_config import ExpeditionConfig, MapConfig
from .gameplay_config import GateType, InstructionConfig, QuantumSimulationConfig, PuzzleConfig, ScoreConfig
from .grammar_config import MapGrammarConfig, PuzzleGrammarConfig, SaveGrammarConfig
from .options import OptionsType, Options, OptionsManager
from .path_config import FileTypes, PathConfig
from .py_cui_config import PyCuiConfig, PyCuiColors
from .test_config import TestConfig
from .visual_config import ColorCode, ColorConfig, HudConfig, PopupConfig, UIConfig, split_text

OptionsManager.init()
UIConfig.init()



from enum import Enum
from typing import Optional, List, Callable

from qrogue.util.config import ColorConfig as CC, InstructionConfig
from .achievements import Unlocks


class _HL:
    indent = "  "

    # objects
    boss = CC.highlight_object("Boss")
    boss_s = CC.highlight_object("Boss's")
    bosses = CC.highlight_object("Bosses")
    circuit = CC.highlight_object("Circuit")
    circuit_matrix = CC.highlight_object("Circuit Matrix")
    collec = CC.highlight_object("Collectibles")
    current_state = CC.highlight_object("Current State")
    door = CC.highlight_object("Door")
    enemies = CC.highlight_object("Enemies")
    energy = CC.highlight_object("Energy")
    gates = CC.highlight_object("Gates")
    gate_i = CC.highlight_object("I Gate")
    gate_cx = CC.highlight_object("CX Gate")
    gate_x = CC.highlight_object("X Gate")
    input_stv = CC.highlight_object("Input State Vector")
    ket_notation = CC.highlight_object("Ket-Notation")
    key = CC.highlight_object("Key")
    keys = CC.highlight_object("Keys")

    mini_boss = CC.highlight_object("MiniBoss")
    output_stv = CC.highlight_object("Output State Vector")
    puzzle = CC.highlight_object("Puzzle")
    puzzle_s = CC.highlight_object("Puzzle's")
    puzzles = CC.highlight_object("Puzzles")
    quantum_circuit = CC.highlight_object("Quantum Circuit")
    quantum_gates = CC.highlight_object("Quantum Gates")
    quantum_state = CC.highlight_object("Quantum State")
    qubit = CC.highlight_object("Qubit")
    qubits = CC.highlight_object("Qubits")
    qubit_s = CC.highlight_object("Qubit(s)")
    riddles = CC.highlight_object("Riddles")
    robot = CC.highlight_object("Robot")
    special_rooms = CC.highlight_object("Special Rooms")
    state = CC.highlight_object("State")
    # state_vectors = CC.highlight_object("StateVectors")
    target_state = CC.highlight_object("Target State")
    target_stv = CC.highlight_object("Target State Vector")

    # actions
    applying = CC.highlight_action("applying")
    buy = CC.highlight_action("buy")
    change = CC.highlight_action("change")
    choose = CC.highlight_action("choose")
    editing = CC.highlight_action("editing")
    edit = CC.highlight_action("edit")
    edits = CC.highlight_action("edits")
    position = CC.highlight_action("position")
    place = CC.highlight_action("place")
    reach = CC.highlight_action("reach")
    select = CC.highlight_action("select")
    talk = CC.highlight_action("talk")
    try_ = CC.highlight_action("try")

    # words
    amplitude = CC.highlight_word("amplitude")
    amplitude_1 = CC.highlight_word("amplitude of 1")
    arbitrary = CC.highlight_word("arbitrary")
    around = CC.highlight_word("around")
    attempt = CC.highlight_word("Attempt")
    bell = CC.highlight_word("Bell, the Master of Entanglement")
    both0 = CC.highlight_word("both 0")
    both1 = CC.highlight_word("both 1")

    cannot = CC.highlight_word("cannot")
    chance_based = CC.highlight_word("chance based")
    collapses = CC.highlight_word("collapses")
    continue_ = CC.highlight_word("Continue")
    details = CC.highlight_word("details")
    deterministic = CC.highlight_word("deterministic")
    difference = CC.highlight_word("Difference")
    display_details = CC.highlight_word("displays details")
    easier = CC.highlight_word("easier")
    edit_attempts = CC.highlight_word("edit attempts")
    enemy = CC.highlight_object("Enemy")
    entanglement = CC.highlight_word("Entanglement")
    equal = CC.highlight_word("equal")
    exam = CC.highlight_word("Exam")
    exit_ = CC.highlight_word("Exit")
    expedition = CC.highlight_word("Expedition")
    finally_ = CC.highlight_word("finally")
    five_parenthesis = CC.highlight_word("5)")
    flee = CC.highlight_word("Flee")
    four = CC.highlight_word("four")
    four_parenthesis = CC.highlight_word("4)")

    gate = CC.highlight_object("Gate")
    give_up = CC.highlight_word("\"Give up\"")
    help_ = CC.highlight_word("Manual")
    hp = CC.highlight_word("don't lose Energy")
    items = CC.highlight_word("Items")
    ket1_symbol = CC.highlight_word("|?>")
    ket2_symbol = CC.highlight_word("|??>")
    ket_symbol_stv01 = CC.highlight_word("|0> 1")
    kronecker = CC.highlight_word("kronecker product")
    leave = CC.highlight_word("\"-Leave-\"")
    list_ = CC.highlight_word("list")
    location = CC.highlight_word("Location")
    locked = CC.highlight_word("locked")
    loses_energy = CC.highlight_word("loses some Energy")
    map_ = CC.highlight_word("map")
    manual = CC.highlight_word("Manual")
    matrix_2x2 = CC.highlight_word("2x2 matrix")
    matrix_4x4 = CC.highlight_word("4x4 matrix")
    mat_vec_mul = CC.highlight_word("matrix-vector multiplication")
    mission = CC.highlight_word("Mission")
    moon = CC.highlight_word("Moon")
    moon_mission = CC.highlight_word("Moon Mission")
    navigate = CC.highlight_word("navigate")
    navigation_panel = CC.highlight_word("Navigation Panel")
    not_zero = CC.highlight_word("not zero")
    number_of_times = CC.highlight_word("number of times")
    often = CC.highlight_word("often")
    one_hp = CC.highlight_word("1 HP")
    one_parenthesis = CC.highlight_word("1)")
    options = CC.highlight_word("Options")
    order = CC.highlight_word("order")
    probability = CC.highlight_word("probability")

    quantum_computing = CC.highlight_word("Quantum Computing")
    quantum_algorithm = CC.highlight_word("Quantum Algorithm")
    qubit_configuration = CC.highlight_word("qubit configuration")
    qubit_configurations = CC.highlight_word("qubit configurations")
    reenter = CC.highlight_word("re-enter")
    removed = CC.highlight_word("removed")
    restart = CC.highlight_word("Restart")
    reward = CC.highlight_word("Reward")
    save = CC.highlight_word("Save")
    serious = CC.highlight_word("serious")
    start_journey = CC.highlight_word("START YOUR JOURNEY")
    state_vector = CC.highlight_word("StateVector")
    state_vectors = CC.highlight_word("StateVectors")
    superposition = CC.highlight_word("Superposition")
    three_parenthesis = CC.highlight_word("3)")
    transforms = CC.highlight_word("transforms")
    two = CC.highlight_word("two")
    two_parenthesis = CC.highlight_word("2)")
    unstable = CC.highlight_word("unstable")
    use = CC.highlight_word("use")
    vanishes = CC.highlight_word("vanishes")
    vanish = CC.highlight_word("vanish")
    win = CC.highlight_word("win")
    zero_energy = CC.highlight_word("0 energy")
    zeros = CC.highlight_word("zeros")
    zero_state = CC.highlight_word("zero-state")

    # keys
    action_keys = CC.highlight_key("Space") + " or " + CC.highlight_key("Enter")
    cancel_keys = CC.highlight_key("Backspace") + " or " + CC.highlight_key("Shift+A") + " or " + \
                  CC.highlight_key("Shift+Left")
    help_keys = CC.highlight_key("H")
    navigation_keys = CC.highlight_key("WASD") + " or " + CC.highlight_key("Arrow Keys")
    pause_keys = CC.highlight_key("P") + " or " + CC.highlight_key("TAB")
    shortcuts = CC.highlight_key("0") + ", " + CC.highlight_key("1") + ", ... , " + CC.highlight_key("9")
    shutdown_keys = CC.highlight_key("CTRL+Q") + " followed by " + CC.highlight_key("Q")

    # tiles
    tile_boss = CC.highlight_tile('B')
    tile_enemy0 = CC.highlight_tile('0')
    tile_mini_boss = CC.highlight_tile('!')


class HelpText(Enum):  # todo: import Controls to replace static key-mentions with dynamic ones?
    Controls = ("controls", "Controls",
        f"Move                  -   {_HL.navigation_keys}\n" \
        f"Navigate menus        -   {_HL.navigation_keys}\n" \
        f"Confirm               -   {_HL.action_keys}\n" \
        f"Cancel/Back           -   {_HL.cancel_keys}\n" \
        f"Scroll in message     -   {_HL.navigation_keys}\n" \
        f"Close message         -   {_HL.action_keys}\n" \
        f"Reopen last message   -   {_HL.help_keys}\n" \
        f"Pause                 -   {_HL.pause_keys}\n" \
        f"Selection shortcuts   -   {_HL.shortcuts}\n" \
        "\n" \
        f"[Should you ever get stuck try to open the pause menu with {_HL.pause_keys} and then select continue. " \
        f"This way the game refocuses and renders again. In case this still doesn't help or doesn't even work you " \
        f"can force-quit the game by pressing {_HL.shutdown_keys}. This will still save everything so it is the " \
        "preferred option over simply closing the window!]")  # let's not mention ESC since it could lead to bugs

    Fight = ("fight", "Puzzles",
        f"Basically {_HL.quantum_computing} is just a lot of complex-valued {_HL.mat_vec_mul}:\n" \
        f"{_HL.circuit_matrix} * {_HL.input_stv} = {_HL.output_stv}\n" \
        f"Your goal is to make the latter {_HL.equal} to the {_HL.target_stv}. While the input and target state " \
        f"vector are given by the {_HL.puzzle}, you can {_HL.change} the {_HL.circuit_matrix} by {_HL.editing} your " \
        f"{_HL.circuit}.\n" \
        f"First you have to {_HL.select} a {_HL.gate} and then {_HL.place} it via {_HL.navigation_keys}. " \
        f"Afterwards use {_HL.action_keys} to confirm the placement. This will update both {_HL.circuit_matrix} " \
        f"and {_HL.output_stv} accordingly. Green numbers in {_HL.output_stv} indicate correct and red incorrect " \
        f"values. Once the whole vector is green, {_HL.output_stv} and {_HL.target_stv} are equal and the " \
        f"{_HL.puzzle} is solved.")

    Ket = ("ket", "Ket-Notation",
        f"The {_HL.ket1_symbol}-notation you can see {_HL.around} the {_HL.state_vectors} and {_HL.circuit_matrix} " \
        f"is called {_HL.ket_notation}. It describes to which {_HL.qubit_configuration} the values correspond to. For " \
        f"example {_HL.ket_symbol_stv01} at ~Target~ means that the target qubit's {_HL.zero_state} should have an " \
        f"{_HL.amplitude_1}, where the amplitude corresponds to a {_HL.probability} of how likely the described " \
        f"{_HL.state} occurs. Let's consider the following (transposed) {_HL.state_vector}:\n" \
        f"|00>  0.707\n" \
        f"|01>  0.707\n" \
        f"|10>    0  \n" \
        f"|11>    0  \n" \
        f"It implies that both {_HL.qubit_configurations} |00> and |01> have an amplitude of 0.707 and therefore " \
        f"correspond to a probability of abs(0.707)^2 = 0.5 = 50%. Since we write the most significant qubit first, " \
        f"this implies that q1=0 is always true and only q0's value comes down to a coin flip.\n" \
        f"In general {_HL.ket_notation} looks like this: |qn...q1q0>, where n is the number of qubits - 1.")

    ParallelGates = ("parallelgates", "Gates in parallel",
        f"Single qubit {_HL.gates} (e.g., X Gate) correspond to {_HL.matrix_2x2} which describe how the {_HL.gate} " \
        f"{_HL.transforms} the two possible inputs (0 and 1). However, {_HL.circuit} with {_HL.two} {_HL.qubits} " \
        f"have {_HL.four} possible inputs (00, 01, 10 and 11) and therefore have to correspond to {_HL.matrix_4x4}. " \
        f"Hence, to extend the original 2x2 matrix of the X Gate we need to calculate a so-called {_HL.kronecker}.\n" \
        f"\n" \
        f"Let's consider a circuit with an {_HL.gate_x} applied to q0. What happens to q1?\n" \
        f"Not performing an operation on it is the same as {_HL.applying} the identity or {_HL.gate_i} (transforming " \
        f"0 to 0 and 1 to 1) to it. Now to calculate the {_HL.kronecker} of X (on q0) and I (on q1) we have to " \
        f"{CC.highlight_word('put X into I')} like this:\n" \
        "    |0>  |1>        |0> |1>\n" \
        "|0> 1*X  0*X  =  |0> X   0 \n" \
        "|1> 0*X  1*X     |1> 0   X \n" \
        "\n" \
        f"and furthermore as we {CC.highlight_word('unpack')} X:\n" \
        "    |00> |01> |10> |11>\n" \
        "|00>  0    1    0    0 \n" \
        "|01>  1    0    0    0 \n" \
        "|10>  0    0    0    1 \n" \
        "|11>  0    0    1    0 \n" \
        "\n" \
        f"The order is important as {CC.highlight_word('putting X into I')} results in a different matrix than " \
        f"{CC.highlight_word('putting I into X')}. But if you look at your {_HL.circuit} you can always go from " \
        f"{CC.highlight_word('top to bottom')}. Just use the {CC.highlight_word('top most')} {_HL.gate} as base and " \
        f"{CC.highlight_word('put the lower')} one(s) into it.")

    SerialGates = ("serialgates", "Gates in series",
        f"Applying two {_HL.gates} in series is as easy as {CC.highlight_word('multiplying')} two matrices.\n" \
        f"Let's consider an {_HL.gate_x} at q0 followed by a {_HL.gate_cx} at q0 (=control), q1 (=target). So we " \
        f"compute a {_HL.matrix_4x4} for the {_HL.gate_x} (see {_HL.kronecker} and then " \
        f"{CC.highlight_word('multiply')} the matrix of the {_HL.gate_cx} {CC.highlight_word('from the left')}:\n" \
        f"    |00> |01> |10> |11>         |00> |01> |10> |11>         |00> |01> |10> |11>\n" \
        f"|00>  1    0    0    0      |00>  0    1    0    0      |00>  0    1    0    0 \n" \
        f"|01>  0    0    0    1   *  |01>  1    0    0    0   =  |01>  0    0    1    0 \n" \
        f"|10>  0    0    1    0      |10>  0    0    0    1      |10>  0    0    0    1 \n" \
        f"|11>  0    1    0    0      |11>  0    0    1    0      |11>  1    0    0    0 \n" \
        f"\n" \
        f"If we take a look at the result's column {CC.highlight_word('|00>')} we can see that the resulting matrix " \
        f"{_HL.transforms} this state to {CC.highlight_word('|11>')} as expected (i.e., X negates q0=0 to q0=1 and " \
        f"then CX negates q1=0 to q1=1 since it's control (q0) is 1).\n" \
        f"Again, in general the {_HL.order} of multiplication matters. At first glance it may seem counter-intuitive " \
        f"that the {_HL.gates} are aligned from {CC.highlight_word('left to right')} in the {_HL.circuit} but have " \
        f"to be {CC.highlight_word('multiplied')} from {CC.highlight_word('right to left')}. But in the " \
        f"{_HL.mat_vec_mul} the {_HL.input_stv} is on the right side while for the {_HL.circuit} the input is on the " \
        f"left. Therefore in both cases simply the gate/matrix {CC.highlight_word('closer')} to the input is " \
        f"considered first.")

    # todo: the split after "collapse to" seems to think that |00> refers to a color code, hence adding // without closing //?
    Amplitudes = ("amplitudes", "Amplitudes",
        f"Let's consider the following {_HL.state_vector}:\n" \
        "|00>  0.707  \n" \
        "|01>  0.707  \n" \
        "|10>    0    \n" \
        "|11>    0    \n" \
        f"From {_HL.ket_notation} we know that both q0=q1=0 and q0=0, q1=1 have an {_HL.amplitude} of 0.707 while " \
        f"it's 0 for the remaining qubit configuration (q0=1, q1=0 and q0=q0=1).\n" \
        f"To transform the amplitude into a probability we have to take it's absolute value squared: abs(0.707)^2 = " \
        f"0.5 which corresponds to 50%. Just as expected this {_HL.state_vector} {_HL.collapses} 50% of the time to " \
        f"|00> and the remaining 50% to |01>. It can never collapse to |10> or |11> because these configurations " \
        f"have an amplitude of 0.\n" \
        f"Since the probabilities need to add up to 1 (100%), the sum of the " \
        f"{CC.highlight_word('squared absolute values')} of a {CC.highlight_object('Quantum StateVector')} also need " \
        f"to be 1.\n" \
        f"In general {_HL.quantum_state}s are described by {CC.highlight_word('complex values')} so unlike in the " \
        f"example above it is very important to not forget to apply the absolute operator before squaring.")

    Riddle = ("ridlle", "Riddles",
        f"{_HL.riddles} are very similar to {_HL.puzzles}, yet a bit different. They also shows you a " \
        f"{_HL.target_state} you need to {_HL.reach} to solve it but you {_HL.cannot} {_HL.try_} it " \
        f"{_HL.arbitrary} {_HL.often}. They have predefined {_HL.number_of_times} you can {_HL.edit} your " \
        f"{_HL.circuit} before they get {_HL.unstable}. Unstable {_HL.riddles} have a 50% chance to {_HL.vanish} " \
        f"together with their {_HL.reward} whenever you update your {_HL.circuit}.")

    Challenge = ("challenge", "Minibosses",
        f"A {_HL.mini_boss} is represented by {_HL.tile_mini_boss} and appears in every {CC.hw('expedition')} and some "
        f"{CC.hw('lessons')}. It has certain properties that {CC.hw('distinguish')} it from a normal {_HL.enemy} or "
        f"{_HL.boss}:\n"
        f"{_HL.indent}1) It will give you a {CC.hw('new')} {_HL.gate} upon challenging it, which you will need to "
        f"{CC.ha('solve')} its {_HL.puzzle}.\n"
        f"{_HL.indent}2) It will only accept your solution if you used a {CC.hw('specified number')} of {_HL.gates}.\n"
        f"{_HL.indent}3) You will {CC.hw('not')} be able to {_HL.flee} from it.\n"
        f"{_HL.indent}4) While its {_HL.puzzle_s} {_HL.target_state} will be a {_HL.zero_state}, the {_HL.input_stv} "
        f"will be {CC.hw('non-zero')}, so you have to kind of think backwards.\n")

    BossFight = ("boss", "Bosses",
        f"A {_HL.boss} is represented by {_HL.tile_boss} and appears in every {CC.hw('expedition')} and some "
        f"{CC.hw('lessons')}. It not only has a {CC.hw('harder')} {_HL.puzzle} than a normal {_HL.enemy} or "
        f"{_HL.mini_boss}, but also additional properties:\n"
        f"{_HL.indent}1) There is a {CC.hw('specified number')} of times you can {_HL.edit} your {_HL.circuit} to "
        f"solve the {_HL.boss_s} {_HL.puzzle}. Should you fail to reach the {_HL.output_stv} before the number of "
        f"remaining edits reaches 0, you lose and have to restart the {CC.hw('level')}. So we advise to not guess "
        f"blindly but really think about your {CC.hw('circuit design')}.\n"
        f"{_HL.indent}2) Harder {_HL.bosses} will have {CC.hw('randomized')} {_HL.input_stv}\n")

    Game = ("game", "About the Game",
        f"QRogue is a game about {_HL.quantum_computing}. You will explore " \
        f"Levels and Expeditions and solve {_HL.puzzles} with the help of {_HL.quantum_gates} to reach even " \
        f"farther places of the quniverse.\n")

    Pause = ("pause", "Pause Menu",
        "In the Pause Menu you can do several things:\n" \
        f"{_HL.continue_} - Leave the Pause Menu and continue where you stopped.\n" \
        f"{_HL.restart} - Restart the current level.\n" \
        f"{_HL.save} - Save your game. Remember that your progress is only saved level-wise.\n" \
        f"{_HL.help_} - If you ever feel stuck and don't remember how certain stuff in the game works open " \
        f"the manual and we will try to help you.\n" \
        f"{_HL.options} - Configure some options of the game, like font size or coloring.\n" \
        f"{_HL.exit_} - Exit the current level.")

    Options = ("options", "Options", "7")   # todo
    Welcome = ("welcome", "Welcome",
        Game[2] + \
        "\nBut before you can explore the universe you have to complete a trainings program.\n" \
        f"Now close this dialog by pressing {_HL.action_keys}. Select {_HL.start_journey} with " \
        f"{_HL.navigation_keys} and confirm your selection with {_HL.action_keys} to begin!")

    LevelSelection = ("levelselection", "Level Selection",
        f"The {CC.hw('Level Selection Screen')} allows you to {CC.ha('revisit')} already completed {CC.hw('Lessons')} "
        f"and go on {CC.hw('specific')} {CC.hw('Expeditions')}. Additionally, you can choose which {CC.ho('Gates')} "
        f"you want to bring along "
        f"for more {CC.hw('experimental freedom')}.\n"
        f"{CC.hw('Beware')} that some {CC.hw('Lessons')} will need {CC.hw('certain')} {CC.ho('Gates')} to be completable. But unless you "
        f"purposefully customize the gate selection, you don't have to worry about it. Finally, you can also specify a "
        f"{CC.hw('seed')} to reduce the randomness of the selected {CC.hw('Level')}.")

    Workbench = ("workbench", "Workbench",
        f"The {CC.hw('Workbench Screen')} allows you to {CC.hw('customize')} your available {CC.ho('Gates')}. You can "
        f"{CC.ha('fuse')} {CC.ho('Gates')} into a new {CC.ho('CombinedGate')} in exchange for {CC.ho('QuantumFusers')} "
        f"or {CC.ha('decompose')} {CC.ho('Gates')} to receive more of said {CC.ho('QuantumFusers')}. Beware that both "
        f"of these actions {CC.hw('destroy')} the original {CC.ho('Gates')}!\n"
        f"For {CC.ha('fusing')}, you first select the {CC.ho('Gates')} you'd like to {CC.ha('fuse')}, and then place them "
        f"however you want in a {CC.hw('Circuit')} - just like you do when solving a {CC.ho('Puzzle')}. The "
        f"{CC.hw('Fusion')} will cost {CC.hw('1')} {CC.ho('QuantumFuser')} for every {CC.ho('Gate')} used. In the end, "
        f"you can give your new {CC.ho('CombinedGate')} a name for easier recognition.\n"
        f"The {CC.hw('naming rules')} are as follows:\n"
        f"{_HL.indent}1) Use between {CC.hw(str(InstructionConfig.COMB_GATE_NAME_MIN_CHARACTERS))} and "
        f"{CC.hw(str(InstructionConfig.COMB_GATE_NAME_MAX_CHARACTERS))} characters\n"
        f"{_HL.indent}2) Only use {CC.hw('letters')} (i.e., no numbers, whitespaces or special characters)\n"
        f"{_HL.indent}3) Must not equal the name of any {CC.hw('Base Gates')} (i.e., any non-CombinedGate in the game)\n"
        f"Lastly, a {CC.ho('CombinedGate')} must {CC.hw('not contain')} any other {CC.ho('CombinedGate')}.")

    def __init__(self, id_: str, name: str, text: str):
        self.__id = id_
        self.__name = name
        self.__text = text

    @property
    def id(self) -> str:
        return self.__id

    @property
    def name(self) -> str:
        return self.__name

    @property
    def text(self) -> str:
        return self.__text


def get_filtered_help_texts(check_unlocks: Callable[[Unlocks], bool]) -> List[HelpText]:
    texts = [HelpText.Game, HelpText.Controls]
    if check_unlocks(Unlocks.ShowEquation):
        texts += [HelpText.Fight, HelpText.Ket, HelpText.ParallelGates, HelpText.SerialGates, HelpText.Amplitudes]
    if check_unlocks(Unlocks.MiniBoss):
        texts.append(HelpText.Challenge)
    if check_unlocks(Unlocks.Boss):
        texts.append(HelpText.BossFight)
    if check_unlocks(Unlocks.LevelSelection):
        texts.append(HelpText.LevelSelection)
    if check_unlocks(Unlocks.Workbench):
        texts.append(HelpText.Workbench)
    return texts


def load_help_text(type_: str) -> Optional[str]:
    for key in HelpText:
        if key.id.lower() == type_.lower():
            return key.text
    return None



from abc import ABC, abstractmethod
from typing import List, Callable, Any, Tuple


class WidgetWrapper(ABC):
    @abstractmethod
    def get_pos(self) -> Tuple[int, int]:
        """
        Get logical position (e.g. grid cell).

        :return: x, y / column, row
        """
        pass

    @abstractmethod
    def get_abs_pos(self) -> Tuple[int, int]:
        """
        Get absolute position (e.g. absolute screen position of top left character).

        :return: x, y / column, row
        """
        pass

    @abstractmethod
    def get_size(self) -> Tuple[int, int]:
        """
        Get logical size (e.g. span of grid cells).

        :return: width, height
        """
        pass

    @abstractmethod
    def get_abs_size(self) -> Tuple[int, int]:
        """
        Get absolute size (e.g. number of characters per row, number of renderable rows).

        :return: width, height
        """
        pass

    @abstractmethod
    def is_selected(self) -> bool:
        pass

    @abstractmethod
    def reposition(self, row: int = None, column: int = None, row_span: int = None, column_span: int = None):
        pass

    @abstractmethod
    def set_title(self, title: str) -> None:
        pass

    @abstractmethod
    def get_title(self) -> str:
        pass

    @abstractmethod
    def add_text_color_rule(self, regex: str, color: int, rule_type: str, match_type: str = 'line',
                            region: List[int] = None, include_whitespace: bool = False, selected_color=None) \
            -> None:
        pass

    @abstractmethod
    def reset_text_color_rules(self) -> None:
        pass

    @abstractmethod
    def activate_individual_coloring(self):
        pass

    @abstractmethod
    def add_key_command(self, keys: List[int], command: Callable[[], Any], overwrite: bool = True) -> Any:
        pass

    @abstractmethod
    def toggle_border(self):
        pass



import os.path
from typing import Optional, Callable

from qrogue.util.config import PathConfig, Config
from qrogue.util.controls import Controls, Keys


class KeyLogger:
    __BUFFER_SIZE = 1024
    __MIN_KEYSTROKES_FOR_FLUSH = 20  # minimum number of keystrokes to log      # todo: make value an Option?

    @staticmethod
    def get_error_marker() -> str:
        return Keys.ErrorMarker.to_char() + Keys.ErrorMarker.to_char() + Keys.ErrorMarker.to_char()

    def __init__(self, write: Optional[Callable[[str, str], None]] = None):
        self.__save_file: Optional[str] = None
        self.__buffer: Optional[str] = None
        self.__keystrokes = 0  # count how many keys are logged to not persist empty or almost empty runs
        self.__is_active = False  # whether log() should log or do nothing

        if write is None:
            self.__write = lambda path, text: PathConfig.write(path, text, may_exist=True, append=True)
        else:
            self.__write = write

    def _is_for_levels(self) -> bool:
        return True

    @property
    def is_initialized(self) -> bool:
        return self.__save_file is not None and self.__buffer is not None

    def set_active(self, active: bool):
        self.__is_active = active

    def _append(self, text: str):
        self.__buffer += text

    def reinit(self, seed: int, level_name: str, save_state: str, save_path: Optional[str] = None,
               activate: bool = True):
        if self.is_initialized:
            # flush the old data if needed
            self.flush_if_useful()
        if activate:  # reactivate if requested
            self.set_active(True)

        # start a new logging session
        if save_path is None:
            self.__save_file = PathConfig.new_key_log_file(seed, self._is_for_levels())
        else:
            self.__save_file = save_path
        self.__buffer = ""
        self._append(level_name)
        self._append("\n")
        self._append(Config.get_log_head(seed))
        self._append(save_state)
        self.__keystrokes = 0

    def log(self, controls: Controls, key_pressed: int, force: bool = False):
        if self.__is_active or force:
            key = controls.encode(key_pressed)
            self._append(key.to_char())
            self.__keystrokes += 1
            self._flush(force=False)

    def flush_if_useful(self):
        """
        Flushes only if the .qrkl-file was already created (meaning we already flushed before) or if the buffer has a
        minimum length, so we don't produce useless files (e.g. immediately quiting a run doesn't provide useful
        information).
        """
        if os.path.exists(self.__save_file) or KeyLogger.__MIN_KEYSTROKES_FOR_FLUSH <= self.__keystrokes:
            self._flush(force=True)

    def _flush(self, force: bool):
        if Config.skip_persisting():
            return
        if force or len(self.__buffer) >= KeyLogger.__BUFFER_SIZE:
            self.__write(self.__save_file, self.__buffer)
            self.__buffer = ""


class OverWorldKeyLogger(KeyLogger):
    def _is_for_levels(self) -> bool:
        return False

    def level_start(self, level_name: str):
        # self._append(
        #    f"{OverWorldKeyLogger.get_level_start_marker()}{level_name}{OverWorldKeyLogger.get_level_start_marker()}"
        # )
        pass


class DummyKeyLogger(OverWorldKeyLogger):
    """
    Acts like a KeyLogger but doesn't actually log/store anything. Can be used to hijack a KeyLogger (e.g., for
    Simulation).
    """

    def __init__(self):
        super().__init__(lambda path, text: None)

    @property
    def is_initialized(self) -> bool:
        return True

    def _append(self, text: str):
        pass

    def reinit(self, seed: int, level_name: str, save_state: str, save_path: Optional[str] = None):
        pass

    def log(self, controls: Controls, key_pressed: int):
        pass

    def flush_if_useful(self):
        pass

    def _flush(self, force: bool):
        pass



from enum import Enum
from typing import Callable, List, Optional

from qrogue.game.logic import actors
from qrogue.game.logic.actors import Controllable, Robot
from qrogue.game.target_factory import EnemyFactory
from qrogue.game.world.navigation import Direction
from qrogue.util import RandomManager, Logger, CheatConfig, PuzzleConfig
from .tiles import Tile, TileCode
from .walk_trigger_tiles import WalkTriggerTile


class Enemy(WalkTriggerTile):
    class _EnemyState(Enum):
        UNDECIDED = 0
        # FREE = 1
        FIGHT = 2
        DEAD = 3
        FLED = 4

    def __init__(self, e_id: int, factory: EnemyFactory,
                 start_fight_callback: Callable[[Robot, "Enemy", Direction], None],
                 get_entangled_tiles_callback: Callable[[int], List[actors.Enemy]],
                 update_entangled_groups_callback: Callable[[actors.Enemy], None], seed: int,
                 tile_id_callback: Callable[[], int] = None):
        super().__init__(TileCode.Enemy)
        self.__rm = RandomManager.create_new(seed)
        self.__factory = factory
        self.__start_fight = start_fight_callback
        self.__get_entangled_tiles = get_entangled_tiles_callback
        self.__update_entangled_groups = update_entangled_groups_callback
        self.__id = e_id
        if tile_id_callback is None:
            self.__next_tile_id = lambda: -1
        else:
            self.__next_tile_id = tile_id_callback
        self.__tile_id = self.__next_tile_id()

        self.__state = Enemy._EnemyState.UNDECIDED
        self.__enemy: Optional[actors.Enemy] = None

    @property
    def data(self) -> int:
        return self.eid

    @property
    def eid(self) -> int:
        return self.__id

    @property
    def _state(self) -> _EnemyState:
        if self.__enemy:
            if self.__state == Enemy._EnemyState.FIGHT and not self.__enemy.is_active:
                self.__state = Enemy._EnemyState.DEAD
                self._explicit_trigger()
        return self.__state

    def is_walkable(self, direction: Direction, controllable: Controllable) -> bool:
        if isinstance(controllable, Robot):
            if controllable.used_capacity > 0:     # make sure that the robot has at least one gate
                return super(Enemy, self).is_walkable(direction, controllable)
            else:
                # noting happens in case the robot doesn't have any gates
                return False
        else:
            Logger.instance().error(f"Error! Non-Robot walked over Enemy: {controllable}", show=False, from_pycui=False)

    def _on_walk(self, direction: Direction, controllable: Controllable) -> bool:
        if isinstance(controllable, Robot):
            robot = controllable
            if self._state == Enemy._EnemyState.UNDECIDED:
                if self.__measure():
                    self.__fight(robot, direction)
                else:
                    Logger.instance().info(f"Enemy #{self.__tile_id} vanished.", from_pycui=False)
            elif self._state == Enemy._EnemyState.FIGHT:
                self.__fight(robot, direction)
        return False

    def get_img(self):
        if self._state == Enemy._EnemyState.DEAD:
            return self._invisible
        elif self._state == Enemy._EnemyState.FLED:
            return self._invisible
        else:
            return str(self.__id)

    def _set_state(self, val: _EnemyState) -> None:
        if self._state == Enemy._EnemyState.UNDECIDED:
            self.__state = val
        else:
            if CheatConfig.did_cheat():  # this is a legal state if we used the "Scared Rabbit" cheat
                return
            Logger.instance().throw(RuntimeError(
                f"Illegal enemy state! Tried to set state to {val} although it is already at {self._state}."
            ))

    def __measure(self):
        # since self will also be part of what get_entangled_tiles() returns, we act upon self via entangled_tiles
        if 0 < self.__id <= 9:
            entangled_tiles = self.__get_entangled_tiles(self.__id)
        else:
            entangled_tiles = [self]

        # first do the randomness check because otherwise cheating could mess with some overall randomness
        if self.__rm.get(msg="Target.__measure()") >= PuzzleConfig.calculate_appearance_chance(self.__id) \
                or CheatConfig.is_scared_rabbit():
            state = Enemy._EnemyState.FLED
        else:
            state = Enemy._EnemyState.FIGHT
        for enemy in entangled_tiles:
            enemy._set_state(state)  # this will also set self.__state

        return self.__state is Enemy._EnemyState.FIGHT

    def __fight(self, robot: Robot, direction: Direction):
        Logger.instance().info(f"Starting fight with #{self.__tile_id}", from_pycui=False)
        if self.__enemy is None:
            if self.__factory.robot_based():
                # old signature based on robot (still needed for LevelMaps)
                self.__enemy = self.__factory.produce_enemy(self.__rm, self.__id, robot)
            else:
                # new signature based on StvDifficulty (needed for ExpeditionMaps)
                self.__enemy = self.__factory.produce_enemy(self.__rm, self.__id, (None, None))
        self.__start_fight(robot, self.__enemy, direction)

    def _copy(self) -> "Tile":
        # todo: check if this should be an *exact* copy including same seeds or not
        enemy = Enemy(self.__id, self.__factory, self.__start_fight, self.__get_entangled_tiles,
                      self.__update_entangled_groups, self.__rm.seed, tile_id_callback=self.__next_tile_id)
        self.__update_entangled_groups(enemy)
        return enemy

    def __str__(self) -> str:
        return f"E({self.__id}|{self._state})"


class Boss(WalkTriggerTile):
    def __init__(self, boss: actors.Boss, on_walk_callback: Callable[[Robot, actors.Boss, Direction], None]):
        super().__init__(TileCode.Boss)
        self.__boss = boss
        self.__on_walk = on_walk_callback
        self.__is_active = True

    @property
    def _is_active(self) -> bool:
        if self.__is_active:
            if not self.__boss.is_active:
                self._explicit_trigger()
                self.__is_active = False
        return self.__is_active

    def _on_walk(self, direction: Direction, controllable: Controllable) -> bool:
        if isinstance(controllable, Robot):
            if self._is_active:
                self.__on_walk(controllable, self.__boss, direction)
            return True
        else:
            Logger.instance().error(f"Non-Robot walked on Boss! controllable = {controllable}", show=False,
                                    from_pycui=False)
            return False

    def get_img(self):
        if self._is_active:
            return "B"
        else:
            return self._invisible

    def _copy(self) -> "Tile":
        # Bosses should not be duplicated in a level anyway, so it doesn't matter if we reference the same actors.Boss
        return Boss(self.__boss, self.__on_walk)



from enum import IntEnum
from typing import Callable, Dict, Optional, Tuple, List, Any, Set

from qrogue.game.logic.actors.controllables.robot import RoboProperties, BaseBot
from qrogue.game.logic.collectibles import GateFactory, Key, instruction, Score, CollectibleType, \
    CollectibleFactory, Instruction
from qrogue.game.target_factory import BossFactory, NewEnemyFactory, ChallengeFactory
from qrogue.game.world import tiles
from qrogue.game.world.dungeon_generator.generator import DungeonGenerator
from qrogue.game.world.dungeon_generator.wave_function_collapse import WFCManager, LearnableRoom
from qrogue.game.world.map import CallbackPack, Hallway, Room, ExpeditionMap
from qrogue.game.world.map.rooms import AreaType, DefinedWildRoom, EmptyRoom, SpawnRoom, BossRoom, TreasureRoom, \
    ChallengeRoom
from qrogue.game.world.navigation import Coordinate, Direction
from qrogue.graphics.popups import Popup
from qrogue.util import Logger, RandomManager, MyRandom, StvDifficulty, GateType, DifficultyType, ExpeditionConfig


class _Code(IntEnum):
    # meta codes
    PriorityMul = -1
    Free = PriorityMul  # best priority for free, unbiased cells
    Blocked = 0  # all before Blocked is free, all after Blocked is already taken
    # room codes
    Spawn = 10
    # Shop = 40
    Riddle = 50
    Challenge = 55
    Boss = 60
    Gate = 70
    Phantom = 80
    Wild = 100
    # hallway codes
    North = 1
    East = 2
    South = 4
    West = 8

    @staticmethod
    def normal_rooms() -> "[_Code]":
        return [_Code.Wild, _Code.Spawn]

    @staticmethod
    def special_rooms() -> "[_Code]":
        return [_Code.Boss, _Code.Challenge]

    @staticmethod
    def get_priority(code: "_Code", inverse: bool) -> float:
        if inverse:
            if code < _Code.Blocked:
                return 20
        elif code > _Code.Blocked:
            if code == _Code.Spawn:
                return 200
            if code == _Code.Wild:
                return 12
            if code in [_Code.Riddle, _Code.Challenge, _Code.Gate]:
                return 5
        return 0

    @staticmethod
    def to_string(code: "_Code", justify: bool = False):
        if justify:
            return _Code.__to_string(code).rjust(4, " ").ljust(5, " ")
        else:
            return _Code.__to_string(code)

    @staticmethod
    def __to_string(code: "_Code"):
        if code < _Code.Blocked:
            str_rep = str(int(code))[1:]
            if len(str_rep) < 4:
                return str_rep
            else:
                return str_rep[0:4]
        if code == _Code.Blocked:
            return "x"
        if code == _Code.Spawn:
            return "S"
        if code == _Code.Riddle:
            return "?"
        if code == _Code.Challenge:
            return "!"
        if code == _Code.Boss:
            return "B"
        if code == _Code.Gate:
            return "G"
        if code == _Code.Wild:
            return "W"
        if code == _Code.Phantom:
            return "%"
        return "#"

    @staticmethod
    def from_enum_string(enum_string: str) -> "_Code":
        if enum_string.startswith("_Code"):
            enum_string = enum_string[enum_string.index(".") + 1:]
        return _Code[enum_string]


class RandomLayoutGenerator:
    __MIN_AREA = 10
    __MIN_NORMAL_ROOMS = 4

    class _RLGException(Exception):
        pass

    def __init__(self, width: int, height: int):
        if width * height < RandomLayoutGenerator.__MIN_AREA:
            Popup.error(f"width={width}, height={height} create a too small grid (minimal grid area = "
                        f"{RandomLayoutGenerator.__MIN_AREA}). Please use larger values!")
        self.__width = width
        self.__height = height
        self.__rama: Optional[MyRandom] = None

        # generate empty map
        self.__map: List[List[_Code]] = []
        self.__spawn_pos: Optional[Coordinate] = None
        self.__normal_rooms: Set[Coordinate] = set()
        self.__hallways: Dict[Coordinate, Dict[Coordinate, tiles.Door]] = {}

        self.__prio_sum = 0
        self.__prio_mean = 0.0

        self.__init_empty_map()

    def __init_empty_map(self):
        # generate empty map
        self.__map = [[_Code.Free] * self.__width for _ in range(self.__height)]
        self.__spawn_pos = None

        self.__normal_rooms.clear()
        self.__hallways.clear()

        self.__prio_sum = self.__width * self.__height
        self.__prio_mean = 1.0

    @property
    def __rm(self) -> MyRandom:
        if self.__rama is None:
            raise RandomLayoutGenerator._RLGException("__rm is None!")
        return self.__rama

    def __get(self, pos: Coordinate) -> _Code:
        return self.__map[pos.y][pos.x]

    def __set(self, pos: Coordinate, code: _Code):
        if code in [_Code.Spawn, _Code.Wild]:
            self.__normal_rooms.add(pos)
        self.__map[pos.y][pos.x] = code

    def __new_prio(self):
        prio_sum = 0
        for y in range(self.__height):
            for x in range(self.__width):
                pos = Coordinate(x, y)
                code = self.__get(pos)
                if code < _Code.Blocked:
                    prio_sum += 1
        self.__prio_sum = prio_sum

    def __is_valid_pos(self, pos: Coordinate) -> bool:
        return 0 <= pos.x < self.__width and 0 <= pos.y < self.__height

    def __is_corner(self, pos: Coordinate) -> (Direction, Direction):
        if pos.x == 0 and pos.y == 0:
            return Direction.East, Direction.South
        if pos.x == 0 and pos.y == self.__height - 1:
            return Direction.East, Direction.North
        if pos.x == self.__width - 1 and pos.y == 0:
            return Direction.West, Direction.South
        if pos.x == self.__width - 1 and pos.y == self.__height - 1:
            return Direction.West, Direction.North
        return None

    def __available_directions(self, pos: Coordinate, allow_wildrooms: bool = False) -> [Direction]:
        directions = []
        # check if going into a direction leaves you still in the grid and on a free spot
        for direction in Direction.values():
            new_pos = pos + direction
            if self.__is_valid_pos(new_pos):
                code = self.__get(new_pos)
                if code < _Code.Blocked or allow_wildrooms and code == _Code.Wild:
                    directions.append(direction)
        return directions

    def __get_neighbors(self, pos: Coordinate, free_spots: bool = False) -> List[Tuple[Direction, _Code, Coordinate]]:
        neighbors = []
        for direction in Direction.values():
            new_pos = pos + direction
            if self.__is_valid_pos(new_pos):
                code = self.__get(new_pos)
                if free_spots:
                    if code < _Code.Blocked:
                        neighbors.append((direction, code, new_pos))
                elif code in _Code.normal_rooms():
                    neighbors.append((direction, code, new_pos))
        return neighbors

    def __random_border(self) -> Optional[Coordinate]:
        directions = Direction.values()
        while len(directions) > 0:
            direction = self.__rm.get_element(directions, remove=True, msg="RandomGen_borderDirection")
            if direction == Direction.North or direction == Direction.South:
                y = (self.__height - 1) * (
                        direction == Direction.South)  # for North this is 0, for South self.__height-1
                xs = list(range(0, self.__width))
                while len(xs) > 0:
                    x = self.__rm.get_element(xs, remove=True, msg="RandomGen_borderX")
                    pos = Coordinate(x=x, y=y)
                    if self.__get(pos) < _Code.Blocked:
                        return pos
            else:
                x = (self.__width - 1) * (direction == Direction.East)  # for North this is 0, for South self.__height-1
                ys = list(range(0, self.__height))
                while len(ys) > 0:
                    y = self.__rm.get_element(ys, remove=True, msg="RandomGen_borderY")
                    pos = Coordinate(x=x, y=y)
                    if self.__get(pos) < _Code.Blocked:
                        return pos
        return None

    def __random_coordinate(self) -> Coordinate:
        val = self.__rm.get(msg="RandomDG_coordinate")
        cur_val = 0.0
        for y in range(self.__height):
            for x in range(self.__width):
                pos = Coordinate(x, y)
                code = self.__get(pos)
                # if there is already something at this position, we continue with the next
                if code >= _Code.Blocked: continue

                code = code * _Code.PriorityMul
                cur_val += code / self.__prio_sum
                if val < cur_val:
                    return pos

        raise RandomLayoutGenerator._RLGException(f"Failed to get a random coordinate (generator.py). Please report "
                                                  f"this error as this should not be possible to occur!")

    def __random_free_wildroom_neighbors(self, num: int = 1)\
            -> List[Tuple[Tuple[Direction, _Code, Coordinate], Coordinate]]:
        rooms: Set[Coordinate] = self.__normal_rooms
        room_prios: Dict[Tuple[Direction, _Code, Coordinate], Tuple[int, Coordinate]] = {}
        max_distance = self.__width + self.__height - 2
        prio_sum = 0
        # calculate the priorities of WR neighbors based on the "isolation" (distance to other WR or SR) of the WR and
        # inverse of the isolation of the neighbor (except its original WR of course)
        for room in rooms:
            min_distance = max_distance
            for other in rooms:
                if room is not other:  # check distance to other WRs and SR except itself
                    min_distance = min(min_distance, Coordinate.distance(room, other))
            neighbors = self.__get_neighbors(room, free_spots=True)
            for neighbor in neighbors:
                _, _, neighbor_pos = neighbor
                neighbor_distance = max_distance
                for other in rooms:
                    if room is not other:  # check distance to other WRs and SR (except the WR we know is a neighbor)
                        neighbor_distance = min(neighbor_distance, Coordinate.distance(neighbor_pos, other))
                # high isolation of room and low isolation of neighbor means high priority
                prio = min_distance * (max_distance - neighbor_distance)
                prio_sum += prio
                room_prios[neighbor] = (prio, room)

        if prio_sum == 0:
            Popup.error(f"Illegal prio_sum in RandomLayoutGenerator!\nThis should not be possible to occur, but aside "
                        f"from the randomness during layout generation this doesn't break anything.")
            prio_sum = -1.0

        picked_rooms: List[Tuple[Tuple[Direction, _Code, Coordinate], Coordinate]] = []
        for i in range(num):
            val = self.__rm.get(msg="RandomDG_WRNeighbors")
            cur_val = 0.0
            for room in room_prios:
                prio, rp = room_prios[room]
                cur_val += prio / prio_sum  # both values have the same sign so their fraction is positive
                if val < cur_val:
                    picked_rooms.append((room, rp))
                    room_prios.pop(room)
                    break
        return picked_rooms

    def __add_hallway(self, room1: Coordinate, room2: Coordinate, door: tiles.Door):
        if room1 in self.__hallways:
            self.__hallways[room1][room2] = door
        else:
            self.__hallways[room1] = {room2: door}
        if room2 in self.__hallways:
            self.__hallways[room2][room1] = door
        else:
            self.__hallways[room2] = {room1: door}

    def __place_wild(self, room: Coordinate, door: tiles.Door):
        pos = room + door.direction
        self.__set(pos, _Code.Wild)
        self.__add_hallway(room, pos, door)

    def __place_special_room(self, code: _Code) -> Coordinate:
        # only Boss rooms are locked
        open_state = tiles.DoorOpenState.KeyLocked if code is _Code.Boss else tiles.DoorOpenState.Closed
        while True:
            try:
                pos = self.__random_border()
                self.__set(pos, code)
                direction = self.__rm.get_element(self.__available_directions(pos, allow_wildrooms=True),
                                                  msg="RandomGen_dirSpecialR")
                if direction is None:
                    self.__set(pos, _Code.Blocked)  # position is not suited for a special room
                else:
                    wild_room = pos + direction
                    if self.__is_corner(wild_room):
                        # if the connected WildRoom is in the corner, we swap position between it and the SpecialRoom
                        self.__set(wild_room, code)
                        self.__place_wild(wild_room, tiles.Door(direction.opposite(), open_state))
                        return wild_room
                    else:
                        self.__place_wild(pos, tiles.Door(direction, open_state))
                        return pos
            except RandomLayoutGenerator._RLGException:
                Popup.error("Unimplemented case happened during layout generation!\nSince this could lead to a faulty "
                            "level, we recommend exiting the level and starting it again. Should this error keep "
                            "occurring:", add_report_note=True)

    def __astar_connect_neighbors(self, visited: set, pos: Coordinate) -> Tuple[Coordinate, bool]:
        """

        :param visited:
        :param pos: the position of a cell that has no Hallways that could lead to the target
        :return:
        """
        relevant_neighbors = []
        # only consider the neighbors we haven't visited yet
        for room in self.__get_neighbors(pos):
            _, _, new_pos = room
            if new_pos not in visited:
                relevant_neighbors.append(room)

        if len(relevant_neighbors) > 0:
            # there are neighbors we can connect to
            room = self.__rm.get_element(relevant_neighbors, msg="RandomGen_astarNeighbor")
            door = tiles.Door(room[0])
            self.__add_hallway(pos, room[2], door)
            return room[2], False
        else:
            if self.__get(pos) in _Code.special_rooms():
                Logger.instance().debug(f"SpecialRoom marked as dead end!", from_pycui=False)
            return pos, True  # we found a dead end

    def __astar(self, visited: Set[Coordinate], pos: Coordinate, target: Coordinate, connect: bool = True) \
            -> Tuple[Optional[Set[Coordinate]], bool]:
        """

        :param visited: Coordinates of all cells we already visited
        :param pos: position/Coordinate of the cell we currently check
        :param target: the cell we try to find
        :param connect: whether we try to create new connection(s) in case we cannot find target
        :return: list of dead ends on the path and False if the target cannot be reached, otherwise True
        """
        dead_ends: Set[Coordinate] = set()
        neighbors = list(self.__hallways[pos].keys())

        if target in neighbors:
            return None, True  # we found the target

        while neighbors:
            room = self.__rm.get_element(neighbors, remove=True, msg="RandomGen_astarRoom")
            if room in visited:
                continue
            visited.add(room)
            ret, success = self.__astar(visited, pos=room, target=target, connect=connect)
            if ret:
                dead_ends.update(ret)
            if success:
                return dead_ends, True

        if connect:
            cur_pos = pos
            while True:
                # we come back from a dead end, so we check if we can connect the current cell with a non-visited neighbor
                coordinate, dead_end = self.__astar_connect_neighbors(visited, cur_pos)
                if dead_end:
                    if len(self.__get_neighbors(coordinate, free_spots=True)) > 0:
                        dead_ends.add(coordinate)
                    return dead_ends, False
                else:
                    visited.add(coordinate)
                    ret, success = self.__astar(visited, pos=coordinate, target=target)
                    if ret is not None:
                        dead_ends.update(ret)
                    if success:
                        return dead_ends, True
        else:
            return dead_ends, False

    def __call_astar(self, visited: Set[Coordinate], start_pos: Coordinate, target_pos: Coordinate,
                     connect: bool = True) -> bool:
        dead_ends, success = self.__astar(visited, start_pos, target=target_pos, connect=connect)
        if success:
            return True
        if connect:
            dead_ends = list(dead_ends)
            while dead_ends:
                dead_end = self.__rm.get_element(dead_ends, remove=True, msg="RandomGen_astarDeadEnd")
                relevant_pos = self.__get_neighbors(dead_end, free_spots=True)
                # and try to place a new room to connect the dead end to the rest
                while relevant_pos:
                    direction, _, new_pos = self.__rm.get_element(relevant_pos, remove=True,
                                                                  msg="RandomGen_astarRelevantPos")
                    self.__place_wild(dead_end, tiles.Door(direction))
                    visited.add(new_pos)
                    if self.__call_astar(visited, start_pos=new_pos, target_pos=target_pos):
                        return True
        return False

    def get_hallway(self, pos: Coordinate) -> Optional[Dict[Coordinate, tiles.Door]]:
        if pos in self.__hallways:
            return self.__hallways[pos]
        return None

    def get_room(self, pos: Coordinate) -> Optional[_Code]:
        if self.__is_valid_pos(pos):
            return self.__get(pos)
        return None

    def generate(self, seed: int, validate: bool = False, debug: bool = False) -> bool:
        Logger.instance().info(f"Generating random layout for seed = {seed}.", from_pycui=False)

        self.__init_empty_map()
        self.__rama = RandomManager.create_new(seed)

        try:
            # place the spawn room
            self.__spawn_pos = self.__random_coordinate()
            self.__set(self.__spawn_pos, _Code.Spawn)

            # special case if SpawnRoom is in a corner
            corner = self.__is_corner(self.__spawn_pos)
            if corner:
                if self.__get(self.__spawn_pos + corner[0]) < _Code.Blocked:
                    self.__place_wild(self.__spawn_pos, tiles.Door(corner[0]))
                if self.__get(self.__spawn_pos + corner[1]) < _Code.Blocked:
                    self.__place_wild(self.__spawn_pos, tiles.Door(corner[1]))

            # place the special rooms
            special_rooms = [
                self.__place_special_room(_Code.Boss),
                self.__place_special_room(_Code.Challenge),
            ]

            # create a locked hallway to spawn_pos-neighboring WildRooms if they lead to SpecialRooms
            # directions = self.__available_directions(self.__spawn_pos, allow_wildrooms=True)
            # for direction in directions:
            #    new_pos = self.__spawn_pos + direction
            #    if self.__get(new_pos) == _Code.Wild:
            #        #wild_directions = self.__available_directions(self.__spawn_pos, allow_wildrooms=True)
            #        door = tiles.Door(direction, locked=True)
            #        self.__add_hallway(self.__spawn_pos, new_pos, door)

            self.__new_prio()
            if len(self.__normal_rooms) < RandomLayoutGenerator.__MIN_NORMAL_ROOMS:
                self.__prio_sum -= (len(self.__normal_rooms) + len(special_rooms))  # subtract now blocked Rooms
                pos = self.__random_coordinate()
                self.__set(pos, _Code.Wild)

            if debug:
                print(self)

            # fill up the rest with a couple of WildRooms
            for num in [3, 2, 1, 1]:
                rooms = self.__random_free_wildroom_neighbors(num)
                for room in rooms:
                    direction, code, new_pos = room[0]
                    pos = room[1]
                    self.__set(new_pos, _Code.Wild)
                    self.__add_hallway(pos, new_pos, tiles.Door(direction))

            # add a connection if the SpawnRoom is not connected to a WildRoom
            if self.__spawn_pos not in self.__hallways:
                neighbors = self.__get_neighbors(self.__spawn_pos)
                if len(neighbors) > 0:
                    direction, _, new_pos = self.__rm.get_element(neighbors, msg="RandomGen_neighbor")
                    self.__add_hallway(self.__spawn_pos, new_pos, tiles.Door(direction))

            success = True
            # as last step, add missing Hallways and WildRooms to connect every SpecialRoom with the SpawnRoom
            rooms = list(special_rooms)
            while len(rooms) > 0:
                room = rooms.pop()
                visited: Set[Coordinate] = set(special_rooms)
                start_pos = list(self.__hallways[room].keys())[0]
                visited.add(start_pos)
                if not self.__call_astar(visited, start_pos, self.__spawn_pos):
                    success = False
                    break

            if not success:
                # try to create a connection to solve the issue
                for room in rooms:
                    start_pos = list(self.__hallways[room].keys())[0]
                    visited = set(special_rooms)
                    dead_ends, success = self.__astar(visited, start_pos, self.__spawn_pos)
                    if not success:
                        return False

            if validate:
                return self.__validate()
            else:
                return True

        except RandomLayoutGenerator._RLGException as ex:
            Logger.instance().error(f"An exception occurred while generating a random layout for seed = {seed}:")
            Logger.instance().throw(ex)
            return False

    def __validate(self) -> bool:
        # first check if all SpecialRooms have exactly 1 connection
        for y in range(self.__height):
            for x in range(self.__width):
                pos = Coordinate(x, y)
                code = self.__get(pos)
                if code in [_Code.Boss, _Code.Challenge, _Code.Riddle, _Code.Gate]:
                    connections = self.__hallways[pos]
                    if len(connections) != 1:
                        return False

        # second check if hallways are symmetric and contain the same doors
        for pos in self.__hallways:
            for neigh in self.__hallways[pos]:
                if pos not in self.__hallways[neigh] or self.__hallways[pos][neigh] != self.__hallways[neigh][pos]:
                    print(self)
                    return False

        # third check if we can reach the spawn room from every room with a hallway
        for pos in self.__hallways:
            visited = set()
            if not self.__call_astar(visited, start_pos=pos, target_pos=self.__spawn_pos, connect=False):
                # now we either found a faulty connection or some "phantom rooms" which are only connected to themselves
                self.__map[pos.y][pos.x] = _Code.Phantom
                new_visited = set()
                # we only have to check immediate neighbors since the neighbors are also checked in the outer loop,
                # therefore, also their neighbors and ultimately any of these "phantom rooms" are checked
                for neigh in self.__hallways[pos]:
                    # see if a neighbor can reach spawn_pos (= neighbor is not a "phantom room" and hence the connection
                    # is faulty)
                    if self.__call_astar(new_visited, start_pos=neigh, target_pos=self.__spawn_pos, connect=False):
                        print(self)
                        return False
        return True

    def __str__(self):
        cell_width = 5
        __empty = "     "
        __row_sep = "_" * (cell_width + 1) * self.__width
        str_rep = " " + __row_sep + "\n"
        for y in range(self.__height):
            rows: List[str] = ["|"] * 3
            for x in range(self.__width):
                pos = Coordinate(x, y)
                if pos in self.__hallways:
                    hws = self.__hallways[pos]
                    if pos + Direction.North in hws:
                        rows[0] += "  ^  "
                    else:
                        rows[0] += __empty
                    code = self.__get(pos)
                    if code < _Code.Blocked:
                        rows[1] += f"{_Code.to_string(code, justify=True)}"
                    else:
                        if pos + Direction.West in hws:
                            rows[1] += " <"
                        else:
                            rows[1] += "  "
                        rows[1] += _Code.to_string(code)
                        if pos + Direction.East in hws:
                            rows[1] += "> "
                        else:
                            rows[1] += "  "
                    if pos + Direction.South in hws:
                        rows[2] += "  ~  "
                    else:
                        rows[2] += __empty
                    for i in range(len(rows)):
                        rows[i] += "|"
                else:
                    rows[0] += __empty + "|"
                    rows[1] += f"{_Code.to_string(self.__get(pos), justify=True)}|"
                    rows[2] += __empty + "|"
            str_rep += "\n".join(rows) + "\n"
            str_rep += " " + __row_sep + " \n"
        return str_rep


class ExpeditionGenerator(DungeonGenerator):
    __MAX_ROOM_GEN_TRIES = 10
    __BLOCKING_WEIGHT = 2
    __INVALID_WEIGHT = 1_000_000

    @staticmethod
    def get_random_gates(available_gates: List[Instruction], difficulty: StvDifficulty, seed: int) \
            -> List[Instruction]:
        """
        :param available_gates: a list of Instructions to choose from; manipulated in-place, so make sure to pass a copy
                                if the list is needed elsewhere
        :param difficulty: the difficulty defines how many Instructions are picked and their needed difficulty score
        :param seed: determines the randomness of picked Instructions
        :returns: a list of randomly picked Instructions adhering to the given StvDifficulty
        """
        rm = RandomManager.create_new(seed)

        def validate(unique_only: bool = False) -> bool:
            val_variety = num_unique_gates >= difficulty.get_absolute_value(DifficultyType.MinGateVariety, 0, 0)
            if unique_only: return val_variety

            val_length = len(gate_list) >= difficulty.get_absolute_value(DifficultyType.MinAvailableGates, 0, 0)
            val_sum = diff_sum >= difficulty.get_absolute_value(DifficultyType.MinGateDifficulty, 0, 0)
            return val_variety and val_length and val_sum

        gate_list = []
        gate_types = set()
        num_unique_gates = 0    # cannot simply check size of gate_types because of CombinedGates
        non_unique_gates = []
        diff_sum = 0
        # first make sure to pick enough unique gates
        while len(available_gates) > 0 and not validate(unique_only=True):
            gate: Instruction = rm.get_element(available_gates, remove=True)
            if gate.gate_type in gate_types:
                non_unique_gates.append(gate)   # save for potential picks later
                continue
            # don't add the type of Combined to the set because usually CombinedGates are not the equal, even though
            #  they share the same GateType (and we don't care enough for the rare case where they are equal)
            if gate.type is not GateType.Combined:
                gate_types.add(gate.gate_type)

            gate_list.append(gate)
            num_unique_gates += 1
            diff_sum += gate.difficulty

        # add back the non-unique gates to be pick-able again
        available_gates += non_unique_gates

        # now make sure to fulfill the remaining criteria
        while len(available_gates) > 0 and not validate():
            # now we can just pick random gates and add them until we fulfill all criteria
            gate: Instruction = rm.get_element(available_gates, remove=True)
            gate_list.append(gate)
            num_unique_gates += 1
            diff_sum += gate.difficulty

        # try to add more gates on top if possible to make it more exciting
        if len(available_gates) > 0:
            # regardless of difficulty, try to add one more for good measure
            gate_list.append(rm.get_element(available_gates, remove=True))

            # for lower difficulties, add more until we have an (optional) minimum number to make it more exciting
            while len(gate_list) < ExpeditionConfig.MIN_NUM_OF_GATES_GOAL and len(available_gates) > 0:
                gate_list.append(rm.get_element(available_gates, remove=True))

        return gate_list

    def __create_enemy(self, enemy_seed: int, enemy_id: int, room_pos: Coordinate, enemy_factory: NewEnemyFactory,
                       enemy_groups_by_room: Dict[Coordinate, Dict[int, List[tiles.Enemy]]]) -> tiles.Enemy:
        enemy: Optional[tiles.Enemy] = None

        def get_entangled_tiles(id_: int) -> List[tiles.Enemy]:
            if room_pos in enemy_groups_by_room:
                room_dic = enemy_groups_by_room[room_pos]
                return room_dic[id_]
            else:
                return [enemy]

        def update_entangled_room_groups(new_enemy: tiles.Enemy):
            if room_pos not in enemy_groups_by_room:
                enemy_groups_by_room[room_pos] = {}
            if enemy_id not in enemy_groups_by_room[room_pos]:
                enemy_groups_by_room[room_pos][enemy_id] = []
            enemy_groups_by_room[room_pos][enemy_id].append(new_enemy)

        enemy = tiles.Enemy(enemy_id, enemy_factory, self.__cbp.start_fight, get_entangled_tiles,
                            update_entangled_room_groups, enemy_seed, self._next_tile_id)
        update_entangled_room_groups(enemy)
        return enemy

    def __init__(self, wfc_manager: WFCManager, check_achievement: Callable[[str], bool],
                 trigger_event: Callable[[str], None], leave_map_callback: Callable[[], None],
                 get_available_gates_callback: Callable[[], List[Instruction]],
                 callback_pack: CallbackPack, width: int = DungeonGenerator.WIDTH,
                 height: int = DungeonGenerator.HEIGHT):
        super(ExpeditionGenerator, self).__init__(width, height)
        self.__check_achievement = check_achievement
        self.__trigger_event = trigger_event
        self.__leave_map = leave_map_callback
        self.__get_available_gates = get_available_gates_callback
        self.__cbp = callback_pack
        self.__next_target_id = 0
        self.__next_tile_id = 0

        self.__wfc_manager = wfc_manager

    def _next_target_id(self) -> int:
        val = self.__next_target_id
        self.__next_target_id += 1
        return val

    def _next_tile_id(self) -> int:
        val = self.__next_tile_id
        self.__next_tile_id += 1
        return val

    def generate(self, map_seed: int, data: Tuple[RoboProperties, StvDifficulty, int]) \
            -> Tuple[Optional[ExpeditionMap], bool]:
        """
        Data tuple consists of:
            - RoboProperties to determine the number of qubits, circuit space and available gates
            - StvDifficulty to determine the difficulty of generated puzzles
            - puzzle seed to determine the randomness of generated puzzles

        :param map_seed: the seed used to create the layout of the ExpeditionMap
        :param data: a tuple describing puzzle generation
        :returns: the generated ExpeditionMap and True if generation was successful, None and False otherwise
        """
        robo_props, difficulty, puzzle_seed = data
        robot = BaseBot.from_properties(robo_props, self.__cbp.game_over)
        if robot.used_capacity <= 0:
            # choose a random subset of globally available gates based on difficulty
            for gate in self.get_random_gates(self.__get_available_gates(), difficulty, puzzle_seed):
                robot.give_collectible(gate)
        else:
            # check validity of gates available to the robot
            val_code, val_data = ExpeditionMap.validate_gates_for_difficulty(difficulty,
                                                                             robot.get_available_instructions())
            if val_code != 0:
                Logger.instance().error("Failed to generate ExpeditionMap based on the given RoboProperties and "
                                        f"StvDifficulty: code={val_code}, data={val_data}", False, from_pycui=False)
                return None, False

        map_rm = RandomManager.create_new(map_seed)         # used for layout, rooms (tile placement), collectibles
        puzzle_rm = RandomManager.create_new(puzzle_seed)   # used for all puzzles

        gate_factory = GateFactory.quantum()
        boss_factory = BossFactory.from_robot(difficulty, robot)
        challenge_factory = ChallengeFactory.from_robot(difficulty, robot)
        typed_collectible_factory: Dict[Optional[CollectibleType], CollectibleFactory] = {
            None: CollectibleFactory([Score(100)]),  # default factory
            CollectibleType.Gate: CollectibleFactory([Score(200)]),
            CollectibleType.Pickup: CollectibleFactory([Score(150)]),
            CollectibleType.Key: CollectibleFactory([Key()]),   # needed for boss-key placement
        }

        def get_collectible_factory(type_: CollectibleType) -> CollectibleFactory:
            if type_ in typed_collectible_factory:
                return typed_collectible_factory[type_]
            return typed_collectible_factory[None]

        main_gate: Instruction = gate_factory.produce(map_rm)
        assert main_gate is not None and isinstance(main_gate, Instruction), f"Invalid product of GateFactory: " \
                                                                             f"\"{main_gate}\"!"
        dungeon_boss = boss_factory.produce(puzzle_rm, include_gates=[main_gate])
        dungeon_challenge = challenge_factory.produce(puzzle_rm, include_gates=[main_gate])

        #####################################################################################################
        # create a variety of enemy factories for normal puzzle generation
        # factories are picked room-wise
        enemy_factories = [
            # plain and simple: one level less CircuitExuberance, 0-state inputs
            NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1,
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions, False),
            # less CircuitExuberance and QubitExuberance, but rotated inputs
            NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1, DifficultyType.QubitExuberance: -1
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions, True),
            # use XGate on inputs - based on QubitExuberance, it should be possible to apply XGate to every qubit
            NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1,
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions,
                [instruction.XGate()] * robot.num_of_qubits),
            # use HGate on inputs
            NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1, DifficultyType.QubitExuberance: -1
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions,
                [instruction.HGate()] * robot.num_of_qubits),
        ]
        enemy_factory_priorities = [0.25, 0.1, 0.35, 0.3]
        # add more factories (varying in input stvs) for higher difficulties
        if difficulty.level > 2:
            # use SGate on inputs
            enemy_factories.append(NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1, DifficultyType.QubitExuberance: -1
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions,
                [instruction.SGate()] * robot.num_of_qubits))
            enemy_factory_priorities.append(0.2)
        if difficulty.level > 3:
            # use SGate on inputs
            enemy_factories.append(NewEnemyFactory(difficulty.create_relative({
                DifficultyType.CircuitExuberance: -1, DifficultyType.QubitExuberance: -1
            }), robot.num_of_qubits, robot.circuit_space, robot.get_available_instructions,
                [instruction.YGate()] * robot.num_of_qubits))
            enemy_factory_priorities.append(0.3)
        enemy_groups_by_room = {}
        #####################################################################################################

        rooms: List[List[Optional[Room]]] = [[None for _ in range(self.width)] for _ in range(self.height)]
        spawn_room = None
        created_hallways = {}
        layout = RandomLayoutGenerator(self.width, self.height)
        if layout.generate(map_seed, validate=True):
            # extract all wild rooms, so we can choose where to place the key
            wild_room_coordinates: List[Coordinate] = []
            for y in range(self.height):
                for x in range(self.width):
                    pos = Coordinate(x, y)
                    code = layout.get_room(pos)
                    if code is _Code.Wild:
                        wild_room_coordinates.append(pos)
            # choose a random wild room to place the key in
            key_room_pos = map_rm.get_element(wild_room_coordinates)

            wild_room_generator = self.__wfc_manager.get_generator(AreaType.WildRoom, map_rm)
            for y in range(self.height):
                for x in range(self.width):
                    pos = Coordinate(x, y)
                    code = layout.get_room(pos)
                    if code is not None and code > _Code.Blocked:
                        room_hallways = {
                            Direction.North: None, Direction.East: None, Direction.South: None, Direction.West: None,
                        }
                        hallways = layout.get_hallway(pos)
                        if hallways is None:
                            if code == _Code.Wild:
                                # it is completely fine if it happens that an isolated WildRoom was generated
                                continue
                            else:
                                Popup.error(f"Found a SpecialRoom ({code}) without connecting Hallways for seed = "
                                            f"{map_seed}.", add_report_note=True)

                        direction: Optional[Direction] = None
                        for neighbor in hallways:
                            direction = Direction.from_coordinates(pos, neighbor)
                            opposite = direction.opposite()
                            # get hallway from neighbor if it exists, otherwise create it
                            if neighbor in created_hallways and opposite in created_hallways[neighbor]:
                                hallway = created_hallways[neighbor][opposite]
                            else:
                                hallway = Hallway(hallways[
                                                      neighbor])  # todo create door here and only check if it should be locked or not?
                                if neighbor in created_hallways:
                                    created_hallways[neighbor][opposite] = hallway
                                else:
                                    created_hallways[neighbor] = {opposite: hallway}

                            # store the hallway so the neighbors can find it if necessary
                            if pos not in created_hallways:
                                created_hallways[pos] = {}
                            created_hallways[pos][direction] = hallway
                            room_hallways[direction] = hallway

                        room = None
                        if code == _Code.Spawn:
                            spawn_room = pos
                            room = SpawnRoom(self.__leave_map,
                                             north_hallway=room_hallways[Direction.North],
                                             east_hallway=room_hallways[Direction.East],
                                             south_hallway=room_hallways[Direction.South],
                                             west_hallway=room_hallways[Direction.West],
                                             place_teleporter=False)
                        elif code == _Code.Wild:
                            enemy_factory = puzzle_rm.get_element_prioritized(enemy_factories, enemy_factory_priorities,
                                                                              msg="RandomDG_elemPrioritized")

                            def tile_from_tile_data(tile_code: tiles.TileCode, tile_data: Any) -> tiles.Tile:
                                if tile_code == tiles.TileCode.Enemy:
                                    enemy_seed = puzzle_rm.get_seed("Create enemy for expedition")
                                    return self.__create_enemy(enemy_seed, tile_data, pos, enemy_factory,
                                                               enemy_groups_by_room)
                                elif tile_code == tiles.TileCode.CollectibleScore:
                                    return tiles.Collectible(Score(tile_data))
                                elif tile_code == tiles.TileCode.Collectible:
                                    return tiles.Collectible(get_collectible_factory(tile_data).produce(map_rm))
                                elif tile_code == tiles.TileCode.Wall:
                                    return tiles.Wall()
                                elif tile_code == tiles.TileCode.Obstacle:
                                    return tiles.Obstacle()
                                else:
                                    return tiles.Floor()

                            # todo: should this be optional based on difficulty/progress (#tunneling)?
                            # fourth check if we can reach every door from every door of this room (= doors are
                            # reachable for player, implies that the previous layout reachability-check still holds)

                            static_entries = {Room.mid(): tiles.Collectible(Key())} if pos == key_room_pos \
                                else None
                            gen_tries = 0
                            while gen_tries < ExpeditionGenerator.__MAX_ROOM_GEN_TRIES:
                                tile_matrix: List[List[LearnableRoom.TileData]] = wild_room_generator.generate(
                                    seed=map_rm.get_seed("generating a room in ExpeditionGenerator"),
                                    static_entries=static_entries
                                )
                                tile_list = [tile_from_tile_data(entry.code, entry.data)
                                             for row in tile_matrix for entry in row]
                                # place the Key again because static_entries can only set a universal Pickup and not
                                #  directly a Key (since we don't want to place Keys at random positions, WFC actually
                                #  does not learn to place Keys, hence, static_entries can only use Pickups to collapse
                                #  the wave functions correctly)
                                if pos == key_room_pos:
                                    tile_list[Room.coordinate_to_index(Room.mid(), in_room=True)] = \
                                        tiles.Collectible(Key())

                                if self.correct_tile_list(tile_list, room_hallways) >= 0:
                                    break
                                Logger.instance().warn(f"Failed to generate a room: try #{gen_tries}, seed={map_seed}")
                                gen_tries += 1
                            if gen_tries >= ExpeditionGenerator.__MAX_ROOM_GEN_TRIES:
                                # todo don't care about potentially impossible Expedition?
                                Logger.instance().error("Failed to validate generated room! Expedition might be "
                                                        "impossible to clear.", from_pycui=False)

                            room = DefinedWildRoom(
                                tile_list,
                                north_hallway=room_hallways[Direction.North],
                                east_hallway=room_hallways[Direction.East],
                                south_hallway=room_hallways[Direction.South],
                                west_hallway=room_hallways[Direction.West]
                            )
                            """
                            room = WildRoom(
                                enemy_factory,
                                chance=map_rm.get(ExpeditionGenerator.__MIN_ENEMY_FACTORY_CHANCE,
                                              ExpeditionGenerator.__MAX_ENEMY_FACTORY_CHANCE,
                                              msg="RandomDG_WRPuzzleDistribution"),
                                north_hallway=room_hallways[Direction.North],
                                east_hallway=room_hallways[Direction.East],
                                south_hallway=room_hallways[Direction.South],
                                west_hallway=room_hallways[Direction.West],
                            )"""
                        elif code == _Code.Phantom:
                            room = EmptyRoom(room_hallways)
                        elif direction is not None:
                            # special rooms have exactly 1 neighbor which is already stored in direction
                            hw = room_hallways[direction]
                            if code == _Code.Gate:
                                room = TreasureRoom(tiles.Collectible(main_gate, force_add=True), hw, direction)
                            elif code == _Code.Challenge:
                                room = ChallengeRoom(hw, direction, tiles.Challenger(dungeon_challenge, main_gate,
                                                                                     self.__cbp.open_challenge,
                                                                                     Popup.generic_info))
                            elif code == _Code.Boss:
                                room = BossRoom(hw, direction, tiles.Boss(dungeon_boss, self.__cbp.start_boss_fight))
                        if room:
                            rooms[y][x] = room
            if spawn_room:
                my_map = ExpeditionMap(map_seed, difficulty, main_gate.gate_type, rooms, robot, spawn_room,
                                       self.__check_achievement, self.__trigger_event)
                return my_map, True
            else:
                return None, False
        else:
            return None, False

    @staticmethod
    def correct_tile_list(tile_list: List[tiles.Tile], hallways: Dict[Direction, Optional[Hallway]]) \
            -> int:
        """
        Corrects a given list of tiles (interpreted as Room.INNER_WIDTH * Room.INNER_HEIGHT tile matrix) such that
        robot is able to reach every Hallway from every other Hallway. This way it is guaranteed that a Room with the
        given tile_list and hallways does not block off access to one of its neighboring Rooms.
        Potentially modifies tile_list by removing Obstacles!

        :param tile_list: a Room's tile matrix in list form (interpreted as concatenation of its rows)
        :param hallways: a dictionary of Directions and the Hallways that connect to other Rooms in this Direction
        :return: how many changes were made on tile_list, with -1 indicating that the tile_list cannot be corrected
        """
        if len(hallways) <= 0:
            return 0

        def is_blocking(t: tiles.Tile) -> bool:
            return t.code in [tiles.TileCode.Obstacle, tiles.TileCode.Wall]

        # gather positions of all possibly blocking tiles
        blocking_tiles: Dict[Coordinate, tiles.Tile] = {}
        for i, tile in enumerate(tile_list):
            if is_blocking(tile):
                blocking_tiles[Room.index_to_coordinate(i)] = tile

        num_of_changes = 0
        # calculate weights for every position
        weight_matrix: Dict[Coordinate, int] = {}
        for x in range(Room.INNER_WIDTH):
            for y in range(Room.INNER_HEIGHT):
                coor = Coordinate(x, y)
                # initialize weight with the minimum distance to one of the hallway entrances
                weight = min([Coordinate.distance(coor, Room.direction_to_hallway_entrance_pos(direction))
                              for direction in hallways.keys()])
                if weight > 0 and is_blocking(tile_list[Room.coordinate_to_index(coor)]):
                    # if weight is 0 (i.e. coor is an entrance) it should stay 0 because it will be a search target
                    weight += ExpeditionGenerator.__BLOCKING_WEIGHT  # increase weight if something is blocking
                weight_matrix[coor] = weight

        def get_weight(position: Coordinate) -> int:
            if position in weight_matrix:
                return weight_matrix[position]
            return ExpeditionGenerator.__INVALID_WEIGHT

        direction: Optional[Direction] = None
        for other_dir in hallways.keys():
            if direction is None:
                direction = other_dir
            elif other_dir is not None:
                start_pos = Room.direction_to_hallway_entrance_pos(direction)
                target_pos = Room.direction_to_hallway_entrance_pos(other_dir)

                # which positions need to be cleared to reach exit from key
                forth = ExpeditionGenerator.path_search(tile_list, is_blocking, get_weight, set(), target_pos,
                                                        start_pos)
                # check the result from searching the other way around
                back = ExpeditionGenerator.path_search(tile_list, is_blocking, get_weight, set(), start_pos,
                                                       target_pos)
                if forth[0] and back[0]:
                    # if both succeeded we use the one that removes less (resulting in a less empty room)
                    if len(forth[1]) < len(back[1]):
                        tiles_to_remove = forth[1]
                    else:
                        tiles_to_remove = back[1]
                elif forth[0]:
                    tiles_to_remove = forth[1]
                elif back[0]:
                    tiles_to_remove = back[1]
                else:
                    # return error code if both directions failed to find a path
                    return -1

                for pos in tiles_to_remove:
                    tile_list[Room.coordinate_to_index(pos)] = tiles.Floor()
                num_of_changes += len(tiles_to_remove)
        return num_of_changes

    @staticmethod
    def path_search(tile_list: List[tiles.Tile], is_blocking: Callable[[tiles.Tile], bool],
                    get_weight: Callable[[Coordinate], int], visited: Set[Coordinate],
                    target_pos: Coordinate, cur_pos: Coordinate) -> Tuple[bool, List[Coordinate]]:
        """
        Searches for a path in tile_list (interpreted as Room.INNER_WIDTH * Room.INNER_HEIGHT tile matrix) from cur_pos
        to target_pos based on positional weights.

        :param tile_list: a Room's tile matrix in list form (interpreted as concatenation of its rows)
        :param is_blocking: tells us whether a given tile is blocking or free to move
        :param get_weight: provides a weight for every room position with lower values being more favourable to move to
        :param visited: set of already visited positions so we don't double check
        :param target_pos: end of the path we search
        :param cur_pos: current position in the path we search
        :return: True and a list of Coordinates of the Obstacles we have to remove to get a valid path, False and a
                 meaningless list if no such path exists (e.g. target is blocked by walls)
        """
        if cur_pos == target_pos:
            return True, []

        # max_ is inclusive
        neighbors = cur_pos.get_neighbors(Coordinate(0, 0), Coordinate(Room.INNER_WIDTH - 1, Room.INNER_HEIGHT - 1))
        neighbors.sort(key=lambda val: get_weight(val))  # continue with the better neighbors first
        for pos in neighbors:
            if pos in visited:
                continue
            visited.add(pos)
            success, tiles_to_remove = ExpeditionGenerator.path_search(tile_list, is_blocking, get_weight,
                                                                       visited, target_pos, pos)
            if success:
                tile = tile_list[pos.x + pos.y * Room.INNER_WIDTH]
                if is_blocking(tile):  # todo theoretically expandable to also take the direction as parameter
                    tiles_to_remove = [pos] + tiles_to_remove
                return True, tiles_to_remove

        return False, []



from typing import List

from qrogue.util.util_functions import to_binary_string, is_power_of_2


def generate_ket(qubit: int, num_of_qubits: int) -> str:
    """
    Generates a str for the ket notation of the given qubit. num_of_qubits is also needed to determine the number of
    leading 0s needed.
    Examples:
        - _generate_ket(1, 2) == "|01>"
        - _generate_ket(0, 1) == "|0>"
        - _generate_ket(3, 4) == "|0011>"

    :param qubit: the id (in terms of MSB-LSB) of the given qubit
    :param num_of_qubits: the number of qubits in the StateVector
    :return: ket notation of the given qubit as str
    """
    return f"|{to_binary_string(qubit, num_of_qubits)}>"


def verify_stv_amplitudes(amplitudes: List[complex], tolerance: float):
    if is_power_of_2(len(amplitudes)):
        amp_sum = sum([c.real ** 2 + c.imag ** 2 for c in amplitudes])
        return 1 - tolerance <= amp_sum <= 1 + tolerance
    return False


def verify_circuit_matrix(matrix: List[List[complex]], tolerance: float) -> bool:
    col_sum: List[float] = [0] * len(matrix)
    for row in matrix:
        row_sum = 0
        for i, val in enumerate(row):
            res_val = abs(val) ** 2
            row_sum += res_val
            col_sum[i] += res_val
        if row_sum < 1 - tolerance or 1 + tolerance < row_sum:
            return False
    for val in col_sum:
        if val < 1 - tolerance or 1 + tolerance < val:
            return False
    return True



from typing import Optional

from qrogue.game.logic.base import StateVector
from qrogue.game.logic.collectibles import Collectible
from qrogue.util import PuzzleConfig
from .target import Target


class Enemy(Target):
    """
    An Enemy is a Target with a certain chance to flee.
    """

    def __init__(self, id_: int, eid: int, target: StateVector, reward: Optional[Collectible],
                 input_: Optional[StateVector] = None):
        """
        Creates an Enemy-Target with a given target state vector and reward.
        :param id_: an integer unique per level to identify the target
        :param eid: enemy id in [0, 9] to calculate certain properties
        :param target: the StateVector to reach
        :param reward: the Collectible to get as a reward
        :param input_: the StateVector to start with (all |0> by default)
        """
        super().__init__(id_, target, reward, input_)
        self.__id: int = eid

    @property
    def flee_energy(self) -> int:
        return PuzzleConfig.calculate_flee_energy(self.__id)

    def flee_check(self) -> bool:
        """
        Check if we are allowed to flee or not.

        :return: True if fleeing was a success, False otherwise
        """
        # todo potentially update docstring? because now we always return True
        return True  # self.__rm.get(msg="Enemy.flee_check()") < PuzzleConfig.calculate_flee_chance(self.__id)

    def __str__(self):
        return "Enemy " + super(Enemy, self).__str__()



import json
from typing import List, Dict, Union, Tuple, Optional

from qrogue.game.world.map.rooms import AreaType
from qrogue.util import PathConfig, MyRandom, Logger
from qrogue.util.util_functions import enum_from_string
from .wfc_generator import WFCGenerator, WFCRoomGenerator


class WFCManager:
    __DEFAULT_FILE = "wfc.txt"
    __AREA_TYPE_START = "_start"
    __AREA_TYPE_END = "_end"
    __POS_WEIGHTS_START = "pos_weights ="
    __TYPE_WEIGHTS_START = "type_weights = "

    def __init__(self):
        self.__area_data: Dict[AreaType, List[WFCRoomGenerator]] = {}

    def learn(self, templates: Optional[List[Tuple[str, bool]]] = None):
        if templates is None:
            templates = WFCRoomGenerator.get_level_list()
        wr_gen = WFCRoomGenerator.from_level_files(templates, AreaType.WildRoom)
        if AreaType.WildRoom in self.__area_data:
            self.__area_data[AreaType.WildRoom].append(wr_gen)
        else:
            self.__area_data[AreaType.WildRoom] = [wr_gen]

    def store(self, file_name: Optional[str] = None):
        if file_name is None: file_name = WFCManager.__DEFAULT_FILE
        data = ""
        for area_type in self.__area_data:
            data += f"{area_type.name}{WFCManager.__AREA_TYPE_START}\n"
            data += f"{len(self.__area_data[area_type])}\n"
            json_strings = []
            for wfc_gen in self.__area_data[area_type]:
                text = f"{WFCManager.__POS_WEIGHTS_START} {WFCGenerator.get_pos_weight_json_string(wfc_gen)}\n"
                text += f"{WFCManager.__TYPE_WEIGHTS_START} {WFCGenerator.get_type_weight_json_string(wfc_gen)}"
                json_strings.append(text)
            data += "; \n".join(json_strings)
            data += "\n"
            data += f"{WFCManager.__AREA_TYPE_END}\n"

        PathConfig.write(file_name, data)

    def load(self, file_name: Optional[str] = None, learn_if_non_existent: Optional[bool] = None):
        if file_name is None: file_name = WFCManager.__DEFAULT_FILE
        if learn_if_non_existent is None: learn_if_non_existent = True
        try:
            data = PathConfig.read(file_name, in_user_path=True)
            line_start = 0
            while line_start < len(data):
                # read area type
                line_end = data.index("\n", line_start)
                area_type = data[line_start:line_end]
                area_type = area_type[:-len(WFCManager.__AREA_TYPE_START)]  # remove meta suffix
                area_type = enum_from_string(AreaType, area_type)

                # read number of upcoming WFCGenerator data (pos and type weights)
                line_start = line_end + 1
                line_end = data.index("\n", line_start)
                num_wfc_gens = int(data[line_start:line_end])

                for i in range(num_wfc_gens):
                    line_start = line_end + 1
                    json_start = line_start + len(WFCManager.__POS_WEIGHTS_START)
                    json_end = data.index(WFCManager.__TYPE_WEIGHTS_START, json_start)
                    json_string = data[json_start:json_end]
                    pos_weights = json.loads(json_string)

                    line_start = json_end + 1
                    json_start = line_start + len(WFCManager.__TYPE_WEIGHTS_START)
                    if i == num_wfc_gens - 1:
                        json_end = data.index(WFCManager.__AREA_TYPE_END, json_start)
                        json_string = data[json_start:json_end]
                        line_end = json_end + len(WFCManager.__AREA_TYPE_END)
                    else:
                        json_end = data.index(WFCManager.__POS_WEIGHTS_START, json_start)
                        json_string = data[json_start:json_end]
                        line_end = json_end
                    type_weights = json.loads(json_string)

                    # convert json dicts to actual dicts (i.e., correct values instead of strings)
                    wfc_gen = WFCRoomGenerator.from_json_dicts(area_type, pos_weights, type_weights)
                    if area_type in self.__area_data:
                        self.__area_data[area_type].append(wfc_gen)
                    else:
                        self.__area_data[area_type] = [wfc_gen]

                line_start = line_end + 1
        except FileNotFoundError as error:
            if learn_if_non_existent:
                # WFC has not been learned yet
                Logger.instance().info(f"WFC file \"{file_name}\" not found. Learning and storing fresh WFC values...",
                                       from_pycui=False)
                self.learn()
                self.store()
            else:
                raise error

    def get_generator(self, area_type: AreaType, selection: Union[int, MyRandom]) -> WFCRoomGenerator:
        if isinstance(selection, int):
            if len(self.__area_data[area_type]) < selection:
                return self.__area_data[area_type][selection]
            else:
                return self.__area_data[area_type][0]   # default value for invalid selection indices
        else:
            return selection.get_element(self.__area_data[area_type], remove=False, msg="WFCManager.get_generator()")



import random
from random import Random
from typing import List, Any

from qrogue.util.config import Config
from qrogue.util.logger import Logger


class MyRandom:
    COUNTER: int = 0
    _MAX_INT: int = 1000000000
    __Next_Id: int = 0

    def __init__(self, seed: int):
        self.__seed = seed % Config.MAX_SEED
        self.__random: Random = Random(seed)
        self.__id: int = MyRandom.__Next_Id
        MyRandom.__Next_Id += 1

    @property
    def seed(self) -> int:
        return self.__seed

    def get(self, min_: float = 0.0, max_: float = 1.0, msg: str = ""):
        MyRandom.COUNTER += 1
        val = min_ + self.__random.random() * (max_ - min_)
        # msg = f"{msg}@{self.__id}_{MyRandom.COUNTER}"
        # PathConfig.write("random_debug.txt", f"{val} | {msg}\n", append=True)
        return val

    def get_int(self, min_: int, max_: int, msg: str = str(COUNTER)) -> int:
        """
        Random integer in the interval [min, max[
        :param msg: a message describing the usage of this method (e.g. caller)
        :param min_: minimum possible int (inclusive)
        :param max_: maximum possible int (exclusive)
        :return: random int in the given range
        """
        return min_ + int(self.get(msg=msg) * (max_ - min_))

    def get_bool(self, msg: str = str(COUNTER)) -> bool:
        return self.get(msg=msg) < 0.5

    def get_seed(self, msg: str = str(COUNTER)) -> int: # todo: introduce class, method, location, sublocation arguments for more consisten messages?
        return self.get_int(min_=0, max_=Config.MAX_SEED, msg=msg)

    def get_element(self, list_: List[Any], remove: bool = False, msg: str = str(COUNTER)) -> Any:
        if len(list_) == 0:
            return None
        index = self.get_int(min_=0, max_=len(list_), msg=msg)
        elem = list_[index]
        if remove:
            try:
                list_.pop(index)
            except ValueError:
                Logger.instance().error(f"{list_} doesn't contain {elem}", show=False, from_pycui=False)
        return elem

    def get_element_prioritized(self, list_: List[Any], priorities: List[float], msg: str = str(COUNTER)) -> Any:
        if len(list_) == 0:
            return None
        if len(list_) == 1:
            return list_[0]
        if len(priorities) == 0:
            return self.get_element(list_, msg=msg)
        if len(priorities) > len(list_):
            # crop priorities if too many values were given
            priorities = priorities[:len(list_)]
        try:
            prio_sum = sum(priorities)
        except:
            Logger.instance().error("No valid priorities provided! Returning a random element to avoid crashing.",
                                    show=False, from_pycui=False)
            return self.get_element(list_, remove=False, msg=msg)
        # elements without a given priority will be less than the minimum given priority by the number of prioritized
        # elements divided by the number of all elements, e.g. [1, 2, 1], 4 elements -> 3/4
        unknown_prio = min(priorities) * len(priorities) / len(list_)

        val = self.get(msg=msg)
        cur_val = 0.0
        index = 0
        for elem in list_:
            if index < len(priorities):
                cur_val += priorities[index] / prio_sum
            else:
                cur_val += unknown_prio / prio_sum
            if val < cur_val:
                return elem
            index += 1
        Logger.instance().throw(ValueError("This line should not be reachable. Please report this error so it can be "
                                           "fixed as soon as possible!"))

    def shuffle_list(self, list_: List[Any]) -> None:
        """
        Shuffles in place.
        """
        self.__random.shuffle(list_)


class RandomManager:
    @staticmethod
    def create_new(seed: int) -> MyRandom:
        if seed is None:
            Logger.instance().throw(Exception("No seed was provided for RandomManager.create_new()!"))
        return MyRandom(seed)



# exporting
from .common_messages import CommonInfos, CommonPopups, CommonQuestions
from .config import *
from .controls import Controls, Keys
from .help_texts import HelpText, load_help_text, get_filtered_help_texts
from .level_info import LevelInfo, LevelData
from .logger import Logger
from .my_random import MyRandom, RandomManager
from .parser_util import ParserErrorListener
from .stv_difficulty import StvDifficulty, DifficultyType

# not exported:
# - KeyLogger etc.
# - GameSimulator
# - achievements
# - util_functions
# - quantum_functions


LevelInfo.init()

# importing
# none
